# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import tempfile
import unittest
from unittest import TestCase

from spdx.parsers.rdf import Parser as RDFParser
from spdx.parsers.rdfbuilders import Builder as RDFBuilder
from spdx.parsers.loggers import StandardLogger
from spdx.parsers.tagvalue import Parser as TVParser
from spdx.parsers.tagvaluebuilders import Builder as TVBuilder
from spdx.parsers.jsonparser import Parser as JSONParser
from spdx.parsers.yamlparser import Parser as YAMLParser
from spdx.parsers.xmlparser import Parser as XMLParser
from spdx.parsers.jsonyamlxmlbuilders import Builder as JSONYAMLXMLBuilder
import spdx.writers.rdf as rdfwriter
import spdx.writers.tagvalue as tvwriter
import spdx.writers.json as jsonwriter
import spdx.writers.yaml as yamlwriter
import spdx.writers.xml as xmlwriter

from tests import utils_test


def get_temp_file(extension=''):
    """
    Return a unique new temporary file location to a non-existing
    temporary file that can safely be created without a risk of name
    collision.
    """

    if extension and not extension.startswith('.'):
            extension = '.' + extension
    file_name = 'temp_file' + extension
    temp_dir = tempfile.mkdtemp()
    return os.path.join(temp_dir, file_name)


class TestConversions(TestCase):
    maxDiff = None

    def parse_rdf_file(self, file_name):
        with open(file_name, mode='rb') as infile:
            rdfparser = RDFParser(RDFBuilder(), StandardLogger())
            doc, error = rdfparser.parse(infile)
        assert not error
        assert doc.validate() == []
        return doc

    def parse_tagvalue_file(self, file_name):
        with open(file_name, mode='r') as infile:
            tvparser = TVParser(TVBuilder(), StandardLogger())
            tvparser.build()
            doc, error = tvparser.parse(infile.read())
        assert not error
        assert doc.validate() == []
        return doc

    def parse_json_file(self, file_name):
        with open(file_name, mode='r') as infile:
            jsonparser = JSONParser(JSONYAMLXMLBuilder(), StandardLogger())
            doc, error = jsonparser.parse(infile)
        assert not error
        assert doc.validate() == []
        return doc

    def parse_yaml_file(self, file_name):
        with open(file_name, mode='r') as infile:
            yamlparser = YAMLParser(JSONYAMLXMLBuilder(), StandardLogger())
            doc, error = yamlparser.parse(infile)
        assert not error
        assert doc.validate() == []
        return doc

    def parse_xml_file(self, file_name):
        with open(file_name, mode='r') as infile:
            xmlparser = XMLParser(JSONYAMLXMLBuilder(), StandardLogger())
            doc, error = xmlparser.parse(infile)        
        assert not error
        assert doc.validate() == []
        return doc

    def write_tagvalue_file(self, document, file_name):
        with codecs.open(file_name, mode='w', encoding='utf-8') as out:
            tvwriter.write_document(document, out)

    def write_rdf_file(self, document, file_name):
        with open(file_name, mode='wb') as out:
            rdfwriter.write_document(document, out)

    def write_json_file(self, document, file_name):
        with open(file_name, mode='w') as out:
            jsonwriter.write_document(document, out)

    def write_yaml_file(self, document, file_name):
        with open(file_name, mode='w') as out:
            yamlwriter.write_document(document, out)

    def write_xml_file(self, document, file_name):
        with open(file_name, mode='w') as out:
            xmlwriter.write_document(document, out)

    def test_tagvalue_rdf(self):
        doc = self.parse_tagvalue_file(utils_test.get_test_loc('formats/SPDXTagExample.tag'))
        filename = get_temp_file('.rdf')
        self.write_rdf_file(doc, filename)
        self.parse_rdf_file(filename)

    def test_json_rdf(self):
        doc = self.parse_json_file(utils_test.get_test_loc('formats/SPDXJsonExample.json'))
        filename = get_temp_file('.rdf')
        self.write_rdf_file(doc, filename)
        self.parse_rdf_file(filename)

    def test_yaml_rdf(self):
        doc = self.parse_yaml_file(utils_test.get_test_loc('formats/SPDXYamlExample.yaml'))
        filename = get_temp_file('.rdf')
        self.write_rdf_file(doc, filename)
        self.parse_rdf_file(filename)

    def test_xml_rdf(self):
        doc = self.parse_xml_file(utils_test.get_test_loc('formats/SPDXXmlExample.xml'))
        filename = get_temp_file('.rdf')
        self.write_rdf_file(doc, filename)
        self.parse_rdf_file(filename)

    def test_rdf_rdf(self):
        doc = self.parse_rdf_file(utils_test.get_test_loc('formats/SPDXRdfExample.rdf'))
        filename = get_temp_file('.rdf')
        self.write_rdf_file(doc, filename)
        self.parse_rdf_file(filename)

    def test_tagvalue_tagvalue(self):
        doc = self.parse_tagvalue_file(utils_test.get_test_loc('formats/SPDXTagExample.tag'))
        filename = get_temp_file('.tag')
        self.write_tagvalue_file(doc, filename)
        self.parse_tagvalue_file(filename)

    @unittest.skip("This fails because we can read/write the mandatory field snippet_byte_range so far only for "
                   "tv-, json-, yaml- or xml-files, not for rdf-files. https://github.com/spdx/tools-python/issues/274")
    def test_rdf_tagvalue(self):
        doc = self.parse_rdf_file(utils_test.get_test_loc('formats/SPDXRdfExample.rdf'))
        filename = get_temp_file('.tag')
        self.write_tagvalue_file(doc, filename)
        self.parse_tagvalue_file(filename)

    def test_json_tagvalue(self):
        doc = self.parse_json_file(utils_test.get_test_loc('formats/SPDXJsonExample.json'))
        filename = get_temp_file('.tag')
        self.write_tagvalue_file(doc, filename)
        self.parse_tagvalue_file(filename)

    def test_yaml_tagvalue(self):
        doc = self.parse_yaml_file(utils_test.get_test_loc('formats/SPDXYamlExample.yaml'))
        filename = get_temp_file('.tag')
        self.write_tagvalue_file(doc, filename)
        self.parse_tagvalue_file(filename)

    def test_xml_tagvalue(self):
        doc = self.parse_xml_file(utils_test.get_test_loc('formats/SPDXXmlExample.xml'))
        filename = get_temp_file('.tag')
        self.write_tagvalue_file(doc, filename)
        self.parse_tagvalue_file(filename)


    def test_tagvalue_json(self):
        doc = self.parse_tagvalue_file(utils_test.get_test_loc('formats/SPDXTagExample.tag'))
        filename = get_temp_file('.json')
        self.write_json_file(doc, filename)
        self.parse_json_file(filename)

    @unittest.skip("This fails because we can read/write the mandatory field snippet_byte_range so far only for "
                   "tv-, json-, yaml- or xml-files, not for rdf-files. https://github.com/spdx/tools-python/issues/274")
    def test_rdf_json(self):
        doc = self.parse_rdf_file(utils_test.get_test_loc('formats/SPDXRdfExample.rdf'))
        filename = get_temp_file('.json')
        self.write_json_file(doc, filename)
        self.parse_json_file(filename)

    def test_yaml_json(self):
        doc = self.parse_yaml_file(utils_test.get_test_loc('formats/SPDXYamlExample.yaml'))
        filename = get_temp_file('.json')
        self.write_json_file(doc, filename)
        self.parse_json_file(filename)

    def test_xml_json(self):
        doc = self.parse_xml_file(utils_test.get_test_loc('formats/SPDXXmlExample.xml'))
        filename = get_temp_file('.json')
        self.write_json_file(doc, filename)
        self.parse_json_file(filename)

    def test_json_json(self):
        doc = self.parse_json_file(utils_test.get_test_loc('formats/SPDXJsonExample.json'))
        filename = get_temp_file('.json')
        self.write_json_file(doc, filename)
        self.parse_json_file(filename)


    def test_tagvalue_yaml(self):
        doc = self.parse_tagvalue_file(utils_test.get_test_loc('formats/SPDXTagExample.tag'))
        filename = get_temp_file('.yaml')
        self.write_yaml_file(doc, filename)
        self.parse_yaml_file(filename)

    @unittest.skip("This fails because we can read/write the mandatory field snippet_byte_range so far only for "
                   "tv-, json-, yaml- or xml-files, not for rdf-files. https://github.com/spdx/tools-python/issues/274")
    def test_rdf_yaml(self):
        doc = self.parse_rdf_file(utils_test.get_test_loc('formats/SPDXRdfExample.rdf'))
        filename = get_temp_file('.yaml')
        self.write_yaml_file(doc, filename)
        self.parse_yaml_file(filename)

    def test_json_yaml(self):
        doc = self.parse_json_file(utils_test.get_test_loc('formats/SPDXJsonExample.json'))
        filename = get_temp_file('.yaml')
        self.write_yaml_file(doc, filename)
        self.parse_yaml_file(filename)

    def test_xml_yaml(self):
        doc = self.parse_xml_file(utils_test.get_test_loc('formats/SPDXXmlExample.xml'))
        filename = get_temp_file('.yaml')
        self.write_yaml_file(doc, filename)
        self.parse_yaml_file(filename)

    def test_yaml_yaml(self):
        doc = self.parse_yaml_file(utils_test.get_test_loc('formats/SPDXYamlExample.yaml'))
        filename = get_temp_file('.yaml')
        self.write_yaml_file(doc, filename)
        self.parse_yaml_file(filename)


    def test_tagvalue_xml(self):
        doc = self.parse_tagvalue_file(utils_test.get_test_loc('formats/SPDXTagExample.tag'))
        filename = get_temp_file('.xml')
        self.write_xml_file(doc, filename)
        self.parse_xml_file(filename)

    @unittest.skip("This fails because we can read/write the mandatory field snippet_byte_range so far only for "
                   "tv-, json-, yaml- or xml-files, not for rdf-files. https://github.com/spdx/tools-python/issues/274")
    def test_rdf_xml(self):
        doc = self.parse_rdf_file(utils_test.get_test_loc('formats/SPDXRdfExample.rdf'))
        filename = get_temp_file('.xml')
        self.write_xml_file(doc, filename)
        self.parse_xml_file(filename)

    def test_json_xml(self):
        doc = self.parse_json_file(utils_test.get_test_loc('formats/SPDXJsonExample.json'))
        filename = get_temp_file('.xml')
        self.write_xml_file(doc, filename)
        self.parse_xml_file(filename)

    def test_yaml_xml(self):
        doc = self.parse_yaml_file(utils_test.get_test_loc('formats/SPDXYamlExample.yaml'))
        filename = get_temp_file('.xml')
        self.write_xml_file(doc, filename)
        self.parse_xml_file(filename)

    def test_xml_xml(self):
        doc = self.parse_xml_file(utils_test.get_test_loc('formats/SPDXXmlExample.xml'))
        filename = get_temp_file('.xml')
        self.write_xml_file(doc, filename)
        self.parse_xml_file(filename)



# Copyright (c) 2022 spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import total_ordering

from spdx import config


@total_ordering
class License(object):
    def __init__(self, full_name, identifier):
        """if one of the argument is None, we try to map as much as possible
        """
        self._full_name = None
        self._identifier = None
        self.set_full_name(full_name)
        self.set_identifier(identifier)

    @classmethod
    def from_identifier(cls, identifier):
        """If identifier exists in config.LICENSE_MAP
        the full_name is retrieved from it. Otherwise
        the full_name is the same as the identifier.
        """
        return cls(None, identifier)

    @classmethod
    def from_full_name(cls, full_name):
        """
        Return a new License for a full_name. If the full_name exists in
        config.LICENSE_MAP the identifier is retrieved from it.
        Otherwise the identifier is the same as the full_name.
        """
        return cls(full_name, None)

    @property
    def url(self):
        return "http://spdx.org/licenses/{0}".format(self.identifier)

    @property
    def full_name(self):
        return self._full_name

    @full_name.setter
    def full_name(self, value):
        self.set_full_name(value)

    def set_full_name(self, value):

        if value is None:
            return
        if self._identifier is None:
            if value in config.LICENSE_MAP:
                self._identifier = config.LICENSE_MAP[value]
            else:
                self._identifier = value
        self._full_name = value

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        self.set_identifier(value)

    def set_identifier(self, value):
        if value is None:
            return
        if self._full_name is None:
            if value in config.LICENSE_MAP:
                self._full_name = config.LICENSE_MAP[value]
            else:
                self._full_name = value

        self._identifier = value


    def __eq__(self, other):
        return (
            isinstance(other, License)
            and self.identifier == other.identifier
            and self.full_name == other.full_name
        )

    def __lt__(self, other):
        return isinstance(other, License) and self.identifier < other.identifier

    def __str__(self):
        return self.identifier

    def __hash__(self):
        return self.identifier.__hash__()


class LicenseConjunction(License):
    """
    A conjunction of two licenses.
    """

    def __init__(self, license_1, license_2):
        self.license_1 = license_1
        self.license_2 = license_2
        super(LicenseConjunction, self).__init__(self.full_name, self.identifier)

    @property
    def full_name(self):
        license_1_complex = type(self.license_1) == LicenseDisjunction
        license_2_complex = type(self.license_2) == LicenseDisjunction

        return "{0} AND {1}".format(
            _add_parens(license_1_complex, self.license_1.full_name),
            _add_parens(license_2_complex, self.license_2.full_name),
        )

    @property
    def identifier(self):
        license_1_complex = type(self.license_1) == LicenseDisjunction
        license_2_complex = type(self.license_2) == LicenseDisjunction

        return "{0} AND {1}".format(
            _add_parens(license_1_complex, self.license_1.identifier),
            _add_parens(license_2_complex, self.license_2.identifier),
        )


class LicenseDisjunction(License):
    """
    A disjunction of two licenses.
    """

    def __init__(self, license_1, license_2):
        self.license_1 = license_1
        self.license_2 = license_2
        super(LicenseDisjunction, self).__init__(self.full_name, self.identifier)

    @property
    def full_name(self):
        license_1_complex = type(self.license_1) == LicenseConjunction
        license_2_complex = type(self.license_2) == LicenseConjunction

        return "{0} OR {1}".format(
            _add_parens(license_1_complex, self.license_1.full_name),
            _add_parens(license_2_complex, self.license_2.full_name),
        )

    @property
    def identifier(self):
        license_1_complex = type(self.license_1) == LicenseConjunction
        license_2_complex = type(self.license_2) == LicenseConjunction

        return "{0} OR {1}".format(
            _add_parens(license_1_complex, self.license_1.identifier),
            _add_parens(license_2_complex, self.license_2.identifier),
        )


@total_ordering
class ExtractedLicense(License):
    """
    Represent an ExtractedLicense with its additional attributes:
    - text: Extracted text, str. Mandatory.
    - cross_ref: list of cross references.
    - comment: license comment, str.
    - full_name: license name. str or utils.NoAssert.
    """

    def __init__(self, identifier):
        super(ExtractedLicense, self).__init__(None, identifier)
        self.text = None
        self.cross_ref = []
        self.comment = None

    def __eq__(self, other):
        return (
            isinstance(other, ExtractedLicense)
            and self.identifier == other.identifier
            and self.full_name == other.full_name
        )

    def __lt__(self, other):
        return (
            isinstance(other, ExtractedLicense) and self.identifier < other.identifier
        )

    def add_xref(self, ref):
        self.cross_ref.append(ref)

    def validate(self, messages):
        if self.text is None:
            messages.append("ExtractedLicense text can not be None")


def _add_parens(required, text):
    """
    Add parens around a license expression if `required` is True, otherwise
    return `text` unmodified.
    """
    return "({})".format(text) if required else text



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from ply import yacc

from spdx import config
from spdx import license
from spdx import utils
from spdx.parsers.builderexceptions import CardinalityError
from spdx.parsers.builderexceptions import OrderError
from spdx.parsers.builderexceptions import SPDXValueError
from spdx.parsers.lexers.tagvalue import Lexer
from spdx.parsers.loggers import ErrorMessages
from spdx import document

ERROR_MESSAGES = {
    "TOOL_VALUE": "Invalid tool value {0} at line: {1}",
    "ORG_VALUE": "Invalid organization value {0} at line: {1}",
    "PERSON_VALUE": "Invalid person value {0} at line: {1}",
    "CREATED_VALUE_TYPE": "Created value must be date in ISO 8601 format, line: {0}",
    "MORE_THAN_ONE": "Only one {0} allowed, extra at line: {1}",
    "CREATOR_COMMENT_VALUE_TYPE": "CreatorComment value must be free form text between <text></text> tags or"
                                  "single line of text, line:{0}",
    "DOC_LICENSE_VALUE": "Invalid DataLicense value '{0}', line:{1} must be CC0-1.0",
    "DOC_LICENSE_VALUE_TYPE": "DataLicense must be CC0-1.0, line: {0}",
    "DOC_VERSION_VALUE": "Invalid SPDXVersion '{0}' must be SPDX-M.N where M and N are numbers. Line: {1}",
    "DOC_VERSION_VALUE_TYPE": "Invalid SPDXVersion value, must be SPDX-M.N where M and N are numbers. Line: {0}",
    "DOC_NAME_VALUE": "DocumentName must be single line of text, line: {0}",
    "DOC_SPDX_ID_VALUE": "Invalid SPDXID value, SPDXID must be SPDXRef-DOCUMENT, line: {0}",
    "LIC_LIST_VER_VALUE": "Invalid LicenseListVersion '{0}', must be of type M.N where M and N are numbers. Line: {1}",
    "LIC_LIST_VER_VALUE_TYPE": "Could not read value after LicenseListVersion-tag. Line{0}",
    "EXT_DOC_REF_VALUE": "ExternalDocumentRef must contain External Document ID, SPDX Document URI and Checksum"
                         "in the standard format, line:{0}.",
    "DOC_COMMENT_VALUE_TYPE": "DocumentComment value must be free form text between <text></text> tags"
                              "or single line of text, line:{0}",
    "DOC_NAMESPACE_VALUE": 'Invalid DocumentNamespace value {0}, must contain a scheme (e.g. "https:") '
                           'and should not contain the "#" delimiter, line:{1}',
    "DOC_NAMESPACE_VALUE_TYPE": 'Invalid DocumentNamespace value, must contain a scheme (e.g. "https:") '
                                'and should not contain the "#" delimiter, line: {0}',
    "REVIEWER_VALUE_TYPE": "Invalid Reviewer value must be a Person, Organization or Tool. Line: {0}",
    "CREATOR_VALUE_TYPE": "Invalid Reviewer value must be a Person, Organization or Tool. Line: {0}",
    "REVIEW_DATE_VALUE_TYPE": "ReviewDate value must be date in ISO 8601 format, line: {0}",
    "REVIEW_COMMENT_VALUE_TYPE": "ReviewComment value must be free form text between <text></text> tags"
                                 "or single line of text, line:{0}",
    "ANNOTATOR_VALUE_TYPE": "Invalid Annotator value must be a Person, Organization or Tool. Line: {0}",
    "ANNOTATION_DATE_VALUE_TYPE": "AnnotationDate value must be date in ISO 8601 format, line: {0}",
    "ANNOTATION_COMMENT_VALUE_TYPE": "AnnotationComment value must be free form text between <text></text> tags"
                                     "or single line of text, line:{0}",
    "ANNOTATION_TYPE_VALUE": 'AnnotationType must be "REVIEW" or "OTHER". Line: {0}',
    "ANNOTATION_SPDX_ID_VALUE": 'SPDXREF must be ["DocumentRef-"[idstring]":"]SPDXID where'
                                '["DocumentRef-"[idstring]":"] is an optional reference to an external SPDX document'
                                'and SPDXID is a unique string containing letters, numbers, ".","-".',
    "A_BEFORE_B": "{0} Can not appear before {1}, line: {2}",
    "PACKAGE_NAME_VALUE": "PackageName must be single line of text, line: {0}",
    "PKG_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string containing '
                         'letters, numbers, ".", "-".',
    "PKG_VERSION_VALUE": "PackageVersion must be single line of text, line: {0}",
    "PKG_FILE_NAME_VALUE": "PackageFileName must be single line of text, line: {0}",
    "PKG_SUPPL_VALUE": "PackageSupplier must be Organization, Person or NOASSERTION, line: {0}",
    "PKG_ORIG_VALUE": "PackageOriginator must be Organization, Person or NOASSERTION, line: {0}",
    "PKG_DOWN_VALUE": "PackageDownloadLocation must be a url or NONE or NOASSERTION, line: {0}",
    "PKG_FILES_ANALYZED_VALUE": "FilesAnalyzed must be a boolean value, line: {0}",
    "PKG_HOME_VALUE": "PackageHomePage must be a url or NONE or NOASSERTION, line: {0}",
    "PKG_SRC_INFO_VALUE": "PackageSourceInfo must be free form text or single line of text, line: {0}",
    "PKG_CHKSUM_VALUE": "PackageChecksum must be a single line of text, line: {0}",
    "PKG_LICS_CONC_VALUE": "PackageLicenseConcluded must be NOASSERTION, NONE, license identifier "
                           "or license list, line: {0}",
    "PKG_LIC_FFILE_VALUE": "PackageLicenseInfoFromFiles must be, line: {0}",
    "PKG_LICS_DECL_VALUE": "PackageLicenseDeclared must be NOASSERTION, NONE, license identifier "
                           "or license list, line: {0}",
    "PKG_LICS_COMMENT_VALUE": "PackageLicenseComments must be free form text or single line of text, line: {0}",
    "PKG_ATTRIBUTION_TEXT_VALUE": "PackageAttributionText must be free form text or single line of text, line: {0}",
    "PKG_SUM_VALUE": "PackageSummary must be free form text or single line of text, line: {0}",
    "PKG_DESC_VALUE": "PackageDescription must be free form text or single line of text, line: {0}",
    "PKG_COMMENT_VALUE": "PackageComment must be free form text or single line of text, line: {0}",
    "PKG_EXT_REF_VALUE": "ExternalRef must contain category, type, and locator in the standard format, line:{0}.",
    "PKG_EXT_REF_COMMENT_VALUE": "ExternalRefComment must be free form text or single line of text, line:{0}",
    "PKG_VERF_CODE_VALUE": "VerificationCode doesn't match verifcode form, line:{0}",
    "PRIMARY_PACKAGE_PURPOSE_VALUE": 'PrimaryPackagePurpose must be one of APPLICATION, FRAMEWORK, LIBRARY, CONTAINER, '
                                     'OPERATING-SYSTEM, DEVICE, FIRMWARE, SOURCE, ARCHIVE, FILE, INSTALL, OTHER',
    "BUILT_DATE_VALUE_TYPE": "Built date value must be date in ISO 8601 format, line: {0}",
    "RELEASE_DATE_VALUE_TYPE": "Release date value must be date in ISO 8601 format, line: {0}",
    "VALID_UNTIL_DATE_VALUE_TYPE": "Valid until date value must be date in ISO 8601 format, line: {0}",
    "FILE_NAME_VALUE": "FileName must be a single line of text, line: {0}",
    "FILE_COMMENT_VALUE": "FileComment must be free form text or single line of text, line:{0}",
    "FILE_TYPE_VALUE": 'FileType must be one of SOURCE, BINARY, ARCHIVE, APPLICATION, AUDIO, IMAGE, TEXT, VIDEO, '
                       'DOCUMENTATION, SPDX, OTHER, line: {0}',
    "FILE_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string containing '
                          'letters, numbers, ".", "-".',
    "FILE_ATTRIBUTION_TEXT_VALUE": "FileAttributionText must be free form text or single line of text, line: {0}",
    "FILE_CHKSUM_VALUE": "FileChecksum must be a single line of text starting with 'SHA1:', line:{0}",
    "FILE_LICS_CONC_VALUE": "LicenseConcluded must be NOASSERTION, NONE, license identifier or license list, line:{0}",
    "FILE_LICS_INFO_VALUE": "LicenseInfoInFile must be NOASSERTION, NONE or license identifier, line: {0}",
    "FILE_LICS_COMMENT_VALUE": "FileLicenseComments must be free form text or single line of text, line: {0}",
    "FILE_CR_TEXT_VALUE": "FileCopyrightText must be one of NOASSERTION, NONE, free form text or single line of text,"
                          "line: {0}",
    "FILE_NOTICE_VALUE": "FileNotice must be free form text or single line of text, line: {0}",
    "FILE_CONTRIB_VALUE": "FileContributor must be a single line, line: {0}",
    "FILE_DEP_VALUE": "FileDependency must be a single line, line: {0}",
    "ART_PRJ_NAME_VALUE": "ArtifactOfProjectName must be a single line, line: {0}",
    "FILE_ART_OPT_ORDER": "ArtificatOfProjectHomePage and ArtificatOfProjectURI must immediately "
                          "follow ArtifactOfProjectName, line: {0}",
    "ART_PRJ_HOME_VALUE": "ArtificatOfProjectHomePage must be a URL or UNKNOWN, line: {0}",
    "ART_PRJ_URI_VALUE": "ArtificatOfProjectURI must be a URI or UNKNOWN, line: {0}",
    "UNKNOWN_TAG": "Found unknown tag : {0} at line: {1}",
    "LICS_ID_VALUE": "LicenseID must start with 'LicenseRef-', line: {0}",
    "LICS_TEXT_VALUE": "ExtractedText must be free form text or single line of text, line: {0}",
    "LICS_NAME_VALE": "LicenseName must be single line of text or NOASSERTION, line: {0}",
    "LICS_COMMENT_VALUE": "LicenseComment must be free form text or single line of text, line: {0}",
    "LICS_CRS_REF_VALUE": "LicenseCrossReference must be uri as single line of text, line: {0}",
    "RELATIONSHIP_VALUE": "Relationship types must be one of the defined types, line: {0}",
    "RELATIONSHIP_COMMENT_VALUE": "RelationshipComment value must be free form text between <text></text> tags "
                                  "or single line of text, line:{0}",
    "PKG_CPY_TEXT_VALUE": "Package copyright text must be free form text or single line of text, line: {0}",
    "SNIP_SPDX_ID_VALUE": 'SPDXID must be "SPDXRef-[idstring]" where [idstring] is a unique string '
                          'containing letters, numbers, ".", "-".',
    "SNIPPET_NAME_VALUE": "SnippetName must be a single line of text, line: {0}",
    "SNIP_COMMENT_VALUE": "SnippetComment must be free form text or single line of text, line: {0}",
    "SNIP_COPYRIGHT_VALUE": "SnippetCopyrightText must be one of NOASSERTION, NONE, "
                            "free form text or single line of text, line: {0}",
    "SNIP_LICS_COMMENT_VALUE": "SnippetLicenseComments must be free form text or single line of text, line: {0}",
    "SNIPPET_ATTRIBUTION_TEXT_VALUE": "SnippetAttributionText must be free form text or single line of text, line: {0}",
    "SNIP_FILE_SPDXID_VALUE": 'SnippetFromFileSPDXID must be ["DocumentRef-"[idstring]":"] SPDXID '
                              "where DocumentRef-[idstring]: is an optional reference to an external"
                              "SPDX Document and SPDXID is a string containing letters, "
                              'numbers, ".", "-".',
    "SNIP_LICS_CONC_VALUE": "SnippetLicenseConcluded must be NOASSERTION, NONE, license identifier "
                            "or license list, line:{0}",
    "SNIP_LICS_INFO_VALUE": "LicenseInfoInSnippet must be NOASSERTION, NONE or license identifier, line: {0}",
}


class Parser(object):
    def __init__(self, builder, logger):
        self.tokens = Lexer.tokens
        self.builder = builder
        self.logger = logger
        self.error = False
        self.license_list_parser = utils.LicenseListParser()
        self.license_list_parser.build(write_tables=0, debug=0)

    def p_start_1(self, p):
        "start : start attrib "
        pass

    def p_start_2(self, p):
        "start : attrib "
        pass

    def p_attrib(self, p):
        """attrib : spdx_version
                  | spdx_id
                  | data_lics
                  | doc_name
                  | ext_doc_ref
                  | doc_comment
                  | doc_namespace
                  | creator
                  | created
                  | creator_comment
                  | locs_list_ver
                  | reviewer
                  | review_date
                  | review_comment
                  | annotator
                  | annotation_date
                  | annotation_comment
                  | annotation_type
                  | annotation_spdx_id
                  | relationship
                  | relationship_comment
                  | package_name
                  | package_version
                  | pkg_down_location
                  | pkg_files_analyzed
                  | pkg_home
                  | pkg_summary
                  | pkg_src_info
                  | pkg_file_name
                  | pkg_supplier
                  | pkg_orig
                  | pkg_chksum
                  | pkg_verif
                  | pkg_desc
                  | pkg_comment
                  | pkg_attribution_text
                  | pkg_lic_decl
                  | pkg_lic_conc
                  | pkg_lic_ff
                  | pkg_lic_comment
                  | pkg_cr_text
                  | pkg_ext_ref
                  | pkg_ext_ref_comment
                  | primary_package_purpose
                  | built_date
                  | release_date
                  | valid_until_date
                  | file_name
                  | file_type
                  | file_chksum
                  | file_conc
                  | file_lics_info
                  | file_cr_text
                  | file_lics_comment
                  | file_attribution_text
                  | file_notice
                  | file_comment
                  | file_contrib
                  | file_dep
                  | file_artifact
                  | snip_spdx_id
                  | snip_name
                  | snip_comment
                  | snippet_attribution_text
                  | snip_cr_text
                  | snip_lic_comment
                  | snip_file_spdx_id
                  | snip_lics_conc
                  | snip_lics_info
                  | snip_byte_range
                  | snip_line_range
                  | extr_lic_id
                  | extr_lic_text
                  | extr_lic_name
                  | lic_xref
                  | lic_comment
                  | unknown_tag
        """
        pass

    def more_than_one_error(self, tag, line):
        self.error = True
        msg = ERROR_MESSAGES["MORE_THAN_ONE"].format(tag, line)
        self.logger.log(msg)

    def order_error(self, first_tag, second_tag, line):
        """Reports an OrderError. Error message will state that
        first_tag came before second_tag.
        """
        self.error = True
        msg = ERROR_MESSAGES["A_BEFORE_B"].format(first_tag, second_tag, line)
        self.logger.log(msg)

    def p_lic_xref_1(self, p):
        """lic_xref : LICS_CRS_REF LINE"""
        try:
            value = p[2]
            self.builder.add_lic_xref(self.document, value)
        except OrderError:
            self.order_error("LicenseCrossReference", "LicenseName", p.lineno(1))

    def p_lic_xref_2(self, p):
        """lic_xref : LICS_CRS_REF error"""
        self.error = True
        msg = ERROR_MESSAGES["LICS_CRS_REF_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_lic_comment_1(self, p):
        """lic_comment : LICS_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_lic_comment(self.document, value)
        except OrderError:
            self.order_error("LicenseComment", "LicenseID", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("LicenseComment", p.lineno(1))

    def p_lic_comment_2(self, p):
        """lic_comment : LICS_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["LICS_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_extr_lic_name_1(self, p):
        """extr_lic_name : LICS_NAME extr_lic_name_value"""
        try:
            self.builder.set_lic_name(self.document, p[2])
        except OrderError:
            self.order_error("LicenseName", "LicenseID", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("LicenseName", p.lineno(1))

    def p_extr_lic_name_2(self, p):
        """extr_lic_name : LICS_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["LICS_NAME_VALE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_extr_lic_name_value_1(self, p):
        """extr_lic_name_value : LINE"""
        p[0] = p[1]

    def p_extr_lic_name_value_2(self, p):
        """extr_lic_name_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_extr_lic_text_1(self, p):
        """extr_lic_text : LICS_TEXT text_or_line"""
        try:
            value = p[2]
            self.builder.set_lic_text(self.document, value)
        except OrderError:
            self.order_error("ExtractedText", "LicenseID", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("ExtractedText", p.lineno(1))

    def p_extr_lic_text_2(self, p):
        """extr_lic_text : LICS_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["LICS_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_extr_lic_id_1(self, p):
        """extr_lic_id : LICS_ID LINE"""
        try:
            value = p[2]
            self.builder.set_lic_id(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["LICS_ID_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_extr_lic_id_2(self, p):
        """extr_lic_id : LICS_ID error"""
        self.error = True
        msg = ERROR_MESSAGES["LICS_ID_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_unknown_tag(self, p):
        """unknown_tag : UNKNOWN_TAG LINE"""
        self.error = True
        msg = ERROR_MESSAGES["UNKNOWN_TAG"].format(p[1], p.lineno(1))
        self.logger.log(msg)

    def p_file_artifact_1(self, p):
        """file_artifact : prj_name_art file_art_rest
                         | prj_name_art
        """
        pass

    def p_file_artifact_2(self, p):
        """file_artifact : prj_name_art error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_ART_OPT_ORDER"].format(p.lineno(2))
        self.logger.log(msg)

    def p_file_art_rest(self, p):
        """file_art_rest : prj_home_art prj_uri_art
                         | prj_uri_art prj_home_art
                         | prj_home_art
                         | prj_uri_art
        """
        pass

    def p_prj_uri_art_1(self, p):
        """prj_uri_art : ART_PRJ_URI UN_KNOWN"""
        try:
            self.builder.set_file_atrificat_of_project(
                self.document, "uri", utils.UnKnown()
            )
        except OrderError:
            self.order_error("ArtificatOfProjectURI", "FileName", p.lineno(1))

    def p_prj_uri_art_2(self, p):
        """prj_uri_art : ART_PRJ_URI LINE"""
        try:
            value = p[2]
            self.builder.set_file_atrificat_of_project(self.document, "uri", value)
        except OrderError:
            self.order_error("ArtificatOfProjectURI", "FileName", p.lineno(1))

    def p_prj_uri_art_3(self, p):
        """prj_uri_art : ART_PRJ_URI error"""
        self.error = True
        msg = ERROR_MESSAGES["ART_PRJ_URI_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_prj_home_art_1(self, p):
        """prj_home_art : ART_PRJ_HOME LINE"""
        try:
            self.builder.set_file_atrificat_of_project(self.document, "home", p[2])
        except OrderError:
            self.order_error("ArtificatOfProjectHomePage", "FileName", p.lineno(1))

    def p_prj_home_art_2(self, p):
        """prj_home_art : ART_PRJ_HOME UN_KNOWN"""
        try:
            self.builder.set_file_atrificat_of_project(
                self.document, "home", utils.UnKnown()
            )
        except OrderError:
            self.order_error("ArtifactOfProjectName", "FileName", p.lineno(1))

    def p_prj_home_art_3(self, p):
        """prj_home_art : ART_PRJ_HOME error"""
        self.error = True
        msg = ERROR_MESSAGES["ART_PRJ_HOME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_prj_name_art_1(self, p):
        """prj_name_art : ART_PRJ_NAME LINE"""
        try:
            value = p[2]
            self.builder.set_file_atrificat_of_project(self.document, "name", value)
        except OrderError:
            self.order_error("ArtifactOfProjectName", "FileName", p.lineno(1))

    def p_prj_name_art_2(self, p):
        """prj_name_art : ART_PRJ_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["ART_PRJ_NAME_VALUE"].format(p.lineno())
        self.logger.log(msg)

    def p_file_dep_1(self, p):
        """file_dep : FILE_DEP LINE"""
        try:
            value = p[2]
            self.builder.add_file_dep(self.document, value)
        except OrderError:
            self.order_error("FileDependency", "FileName", p.lineno(1))

    def p_file_dep_2(self, p):
        """file_dep : FILE_DEP error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_DEP_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_contrib_1(self, p):
        """file_contrib : FILE_CONTRIB LINE"""
        try:
            value = p[2]
            self.builder.add_file_contribution(self.document, value)
        except OrderError:
            self.order_error("FileContributor", "FileName", p.lineno(1))

    def p_file_contrib_2(self, p):
        """file_contrib : FILE_CONTRIB error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_CONTRIB_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_notice_1(self, p):
        """file_notice : FILE_NOTICE text_or_line"""
        try:
            value = p[2]
            self.builder.set_file_notice(self.document, value)
        except OrderError:
            self.order_error("FileNotice", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("FileNotice", p.lineno(1))

    def p_file_notice_2(self, p):
        """file_notice : FILE_NOTICE error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_NOTICE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_cr_text_1(self, p):
        """file_cr_text : FILE_CR_TEXT file_cr_value"""
        try:
            self.builder.set_file_copyright(self.document, p[2])
        except OrderError:
            self.order_error("FileCopyrightText", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("FileCopyrightText", p.lineno(1))

    def p_file_cr_text_2(self, p):
        """file_cr_text : FILE_CR_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_CR_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_cr_value_1(self, p):
        """file_cr_value : text_or_line"""
        p[0] = p[1]

    def p_file_cr_value_2(self, p):
        """file_cr_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_file_cr_value_3(self, p):
        """file_cr_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_file_lics_comment_1(self, p):
        """file_lics_comment : FILE_LICS_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_file_license_comment(self.document, value)
        except OrderError:
            self.order_error("LicenseComments", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("LicenseComments", p.lineno(1))

    def p_file_lics_comment_2(self, p):
        """file_lics_comment : FILE_LICS_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_LICS_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_attribution_text_1(self, p):
        """file_attribution_text : FILE_ATTRIBUTION_TEXT text_or_line"""
        try:
            value = p[2]
            self.builder.set_file_attribution_text(self.document, value)
        except CardinalityError:
            self.more_than_one_error("FileAttributionText", p.lineno(1))
        except OrderError:
            self.order_error("FileAttributionText", "FileAttributionText", p.lineno(1))

    def p_file_attribution_text_2(self, p):
        """file_attribution_text : FILE_ATTRIBUTION_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_ATTRIBUTION_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_lics_info_1(self, p):
        """file_lics_info : FILE_LICS_INFO file_lic_info_value"""
        try:
            self.builder.set_file_license_in_file(self.document, p[2])
        except OrderError:
            self.order_error("LicenseInfoInFile", "FileName", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["FILE_LICS_INFO_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_file_lics_info_2(self, p):
        """file_lics_info : FILE_LICS_INFO error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_LICS_INFO_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_lic_info_value_1(self, p):
        """file_lic_info_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_file_lic_info_value_2(self, p):
        """file_lic_info_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    # License Identifier
    def p_file_lic_info_value_3(self, p):
        """file_lic_info_value : LINE"""
        value = p[1]
        p[0] = license.License.from_identifier(value)

    def p_conc_license_1(self, p):
        """conc_license : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_conc_license_2(self, p):
        """conc_license : NONE"""
        p[0] = utils.SPDXNone()

    def p_conc_license_3(self, p):
        """conc_license : LINE"""
        value = p[1]
        ref_re = re.compile("LicenseRef-.+", re.UNICODE)
        if (p[1] in config.LICENSE_MAP.keys()) or (ref_re.match(p[1]) is not None):
            p[0] = license.License.from_identifier(value)
        else:
            p[0] = self.license_list_parser.parse(value)

    def p_file_name_1(self, p):
        """file_name : FILE_NAME LINE"""
        try:
            value = p[2]
            self.builder.set_file_name(self.document, value)
            self.builder.set_current_file_name(value)
            self.builder.set_current_file_id(None)

        except OrderError:
            self.order_error("FileName", "PackageName", p.lineno(1))

    def p_file_name_2(self, p):
        """file_name : FILE_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_NAME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_spdx_id(self, p):
        """spdx_id : SPDX_ID LINE"""
        value = p[2]
        try:
            # first parsed spdx id belongs to the document
            if not self.builder.doc_spdx_id_set:
                self.builder.set_doc_spdx_id(self.document, value)

            # else if a package is in scope that doesn't have an id yet, the parsed spdx id belongs to the package
            elif self.builder.current_package_has_name() \
                    and not self.builder.current_package_has_id():
                self.builder.set_pkg_spdx_id(self.document, value)
                self.builder.set_current_package_id(value)

            # else if a file is in scope that doesn't have an id yet, the parsed spdx id belongs to the file
            elif self.builder.current_file_has_name() \
                    and not self.builder.current_file_has_id():
                self.builder.set_file_spdx_id(self.document, value)
                self.builder.set_current_file_id(value)
                if self.builder.has_current_package():
                    self.builder.add_relationship(self.document,
                                                  self.builder.current_package["spdx_id"] + " CONTAINS " + value)
            else:
                raise SPDXValueError("SPDX ID couldn't be assigned properly.  Line no. {0}")
        except SPDXValueError as err:
            self.error = True
            self.logger.log(err.msg.format(p.lineno(2)))

    def p_file_comment_1(self, p):
        """file_comment : FILE_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_file_comment(self.document, value)
        except OrderError:
            self.order_error("FileComment", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("FileComment", p.lineno(1))

    def p_file_comment_2(self, p):
        """file_comment : FILE_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_type_1(self, p):
        """file_type : FILE_TYPE file_type_value"""
        try:
            self.builder.set_file_type(self.document, p[2])
        except OrderError:
            self.order_error("FileType", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("FileType", p.lineno(1))

    def p_file_type_2(self, p):
        """file_type : FILE_TYPE error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_TYPE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_chksum_1(self, p):
        """file_chksum : FILE_CHKSUM CHKSUM"""
        try:
            value = p[2]
            self.builder.set_file_checksum(self.document, value)
        except OrderError:
            self.order_error("FileChecksum", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("FileChecksum", p.lineno(1))

    def p_file_chksum_2(self, p):
        """file_chksum : FILE_CHKSUM error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_CHKSUM_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_conc_1(self, p):
        """file_conc : FILE_LICS_CONC conc_license"""
        try:
            self.builder.set_concluded_license(self.document, p[2])
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["FILE_LICS_CONC_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except OrderError:
            self.order_error("LicenseConcluded", "FileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("LicenseConcluded", p.lineno(1))

    def p_file_conc_2(self, p):
        """file_conc : FILE_LICS_CONC error"""
        self.error = True
        msg = ERROR_MESSAGES["FILE_LICS_CONC_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_file_type_value(self, p):
        """file_type_value : SOURCE
                           | BINARY
                           | ARCHIVE
                           | APPLICATION
                           | AUDIO
                           | IMAGE
                           | FILETYPE_TEXT
                           | VIDEO
                           | DOCUMENTATION
                           | SPDX
                           | OTHER
        """
        p[0] = p[1]

    def p_annotation_type_value(self, p):
        """annotation_type_value : OTHER
                           | REVIEW
        """
        p[0] = p[1]

    def p_pkg_desc_1(self, p):
        """pkg_desc : PKG_DESC text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_desc(self.document, value)
        except CardinalityError:
            self.more_than_one_error("PackageDescription", p.lineno(1))
        except OrderError:
            self.order_error("PackageDescription", "PackageFileName", p.lineno(1))

    def p_pkg_desc_2(self, p):
        """pkg_desc : PKG_DESC error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_DESC_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_comment_1(self, p):
        """pkg_comment : PKG_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("PackageComment", p.lineno(1))
        except OrderError:
            self.order_error("PackageComment", "PackageFileName", p.lineno(1))

    def p_pkg_comment_2(self, p):
        """pkg_comment : PKG_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_attribution_text_1(self, p):
        """pkg_attribution_text : PKG_ATTRIBUTION_TEXT text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_attribution_text(self.document, value)
        except CardinalityError:
            self.more_than_one_error("PackageAttributionText", p.lineno(1))
        except OrderError:
            self.order_error(
                "PackageAttributionText", "PackageAttributionText", p.lineno(1)
            )

    def p_pkg_attribution_text_2(self, p):
        """pkg_attribution_text : PKG_ATTRIBUTION_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_ATTRIBUTION_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_summary_1(self, p):
        """pkg_summary : PKG_SUM text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_summary(self.document, value)
        except OrderError:
            self.order_error("PackageSummary", "PackageFileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageSummary", p.lineno(1))

    def p_pkg_summary_2(self, p):
        """pkg_summary : PKG_SUM error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_SUM_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_cr_text_1(self, p):
        """pkg_cr_text : PKG_CPY_TEXT pkg_cr_text_value"""
        try:
            self.builder.set_pkg_cr_text(self.document, p[2])
        except OrderError:
            self.order_error("PackageCopyrightText", "PackageFileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageCopyrightText", p.lineno(1))

    def p_pkg_cr_text_2(self, p):
        """pkg_cr_text : PKG_CPY_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_CPY_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_ext_refs_1(self, p):
        """pkg_ext_ref : PKG_EXT_REF LINE"""
        try:
            pkg_ext_info = p[2]
            if len(pkg_ext_info.split()) != 3:
                raise SPDXValueError(ERROR_MESSAGES["PKG_EXT_REF_VALUE"].format(p.lineno(2)))
            else:
                pkg_ext_category, pkg_ext_type, pkg_ext_locator = pkg_ext_info.split()
            self.builder.add_pkg_ext_refs(
                self.document, pkg_ext_category, pkg_ext_type, pkg_ext_locator
            )
        except SPDXValueError as err:
            self.error = True
            self.logger.log(err.msg)

    def p_pkg_ext_refs_2(self, p):
        """pkg_ext_ref : PKG_EXT_REF error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_EXT_REF_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_ext_ref_comment_1(self, p):
        """pkg_ext_ref_comment : PKG_EXT_REF_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.add_pkg_ext_ref_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("ExternalRefComment", p.lineno(1))

    def p_pkg_ext_ref_comment_2(self, p):
        """pkg_ext_ref_comment : PKG_EXT_REF_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_EXT_REF_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_cr_text_value_1(self, p):
        """pkg_cr_text_value : text_or_line"""
        p[0] = p[1]

    def p_pkg_cr_text_value_2(self, p):
        """pkg_cr_text_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_pkg_cr_text_value_3(self, p):
        """pkg_cr_text_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_pkg_lic_comment_1(self, p):
        """pkg_lic_comment : PKG_LICS_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_license_comment(self.document, value)
        except OrderError:
            self.order_error("PackageLicenseComments", "PackageFileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageLicenseComments", p.lineno(1))

    def p_pkg_lic_comment_2(self, p):
        """pkg_lic_comment : PKG_LICS_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_LICS_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_lic_decl_1(self, p):
        """pkg_lic_decl : PKG_LICS_DECL conc_license"""
        try:
            self.builder.set_pkg_license_declared(self.document, p[2])
        except OrderError:
            self.order_error("PackageLicenseDeclared", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageLicenseDeclared", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_LICS_DECL_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_lic_decl_2(self, p):
        """pkg_lic_decl : PKG_LICS_DECL error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_LICS_DECL_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_lic_ff_1(self, p):
        """pkg_lic_ff : PKG_LICS_FFILE pkg_lic_ff_value"""
        try:
            self.builder.set_pkg_license_from_file(self.document, p[2])
        except OrderError:
            self.order_error("PackageLicenseInfoFromFiles", "PackageName", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_LIC_FFILE_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_lic_ff_value_1(self, p):
        """pkg_lic_ff_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_pkg_lic_ff_value_2(self, p):
        """pkg_lic_ff_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_pkg_lic_ff_value_3(self, p):
        """pkg_lic_ff_value : LINE"""
        value = p[1]
        p[0] = license.License.from_identifier(value)

    def p_pkg_lic_ff_2(self, p):
        """pkg_lic_ff : PKG_LICS_FFILE error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_LIC_FFILE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_lic_conc_1(self, p):
        """pkg_lic_conc : PKG_LICS_CONC conc_license"""
        try:
            self.builder.set_pkg_licenses_concluded(self.document, p[2])
        except CardinalityError:
            self.more_than_one_error("PackageLicenseConcluded", p.lineno(1))
        except OrderError:
            self.order_error("PackageLicenseConcluded", "PackageFileName", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_LICS_CONC_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_lic_conc_2(self, p):
        """pkg_lic_conc : PKG_LICS_CONC error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_LICS_CONC_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_src_info_1(self, p):
        """pkg_src_info : PKG_SRC_INFO text_or_line"""
        try:
            value = p[2]
            self.builder.set_pkg_source_info(self.document, value)
        except CardinalityError:
            self.more_than_one_error("PackageSourceInfo", p.lineno(1))
        except OrderError:
            self.order_error("PackageSourceInfo", "PackageFileName", p.lineno(1))

    def p_pkg_src_info_2(self, p):
        """pkg_src_info : PKG_SRC_INFO error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_SRC_INFO_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_chksum_1(self, p):
        """pkg_chksum : PKG_CHKSUM CHKSUM"""
        try:
            value = p[2]
            self.builder.set_pkg_checksum(self.document, value)
        except OrderError:
            self.order_error("PackageChecksum", "PackageFileName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageChecksum", p.lineno(1))

    def p_pkg_chksum_2(self, p):
        """pkg_chksum : PKG_CHKSUM error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_CHKSUM_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_verif_1(self, p):
        """pkg_verif : PKG_VERF_CODE LINE"""
        try:
            value = p[2]
            self.builder.set_pkg_verif_code(self.document, value)
        except OrderError:
            self.order_error("PackageVerificationCode", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageVerificationCode", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_VERF_CODE_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_verif_2(self, p):
        """pkg_verif : PKG_VERF_CODE error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_VERF_CODE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_home_1(self, p):
        """pkg_home : PKG_HOME pkg_home_value"""
        try:
            self.builder.set_pkg_home(self.document, p[2])
        except OrderError:
            self.order_error("PackageHomePage", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageHomePage", p.lineno(1))

    def p_pkg_home_2(self, p):
        """pkg_home : PKG_HOME error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_HOME_VALUE"]
        self.logger.log(msg)

    def p_pkg_home_value_1(self, p):
        """pkg_home_value : LINE"""
        p[0] = p[1]

    def p_pkg_home_value_2(self, p):
        """pkg_home_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_pkg_home_value_3(self, p):
        """pkg_home_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_pkg_down_location_1(self, p):
        """pkg_down_location : PKG_DOWN pkg_down_value"""
        try:
            self.builder.set_pkg_down_location(self.document, p[2])
        except OrderError:
            self.order_error("PackageDownloadLocation", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageDownloadLocation", p.lineno(1))

    def p_pkg_down_location_2(self, p):
        """pkg_down_location : PKG_DOWN error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_DOWN_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_files_analyzed_1(self, p):
        """pkg_files_analyzed : PKG_FILES_ANALYZED LINE"""
        try:
            value = p[2]
            self.builder.set_pkg_files_analyzed(self.document, value)
        except CardinalityError:
            self.more_than_one_error("FilesAnalyzed", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_FILES_ANALYZED_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_files_analyzed_2(self, p):
        """pkg_files_analyzed : PKG_FILES_ANALYZED error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_FILES_ANALYZED_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_down_value_1(self, p):
        """pkg_down_value : LINE """
        p[0] = p[1]

    def p_pkg_down_value_2(self, p):
        """pkg_down_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_pkg_down_value_3(self, p):
        """pkg_down_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_pkg_orig_1(self, p):
        """pkg_orig : PKG_ORIG pkg_supplier_values"""
        try:
            self.builder.set_pkg_originator(self.document, p[2])
        except OrderError:
            self.order_error("PackageOriginator", "PackageName", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_ORIG_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("PackageOriginator", p.lineno(1))

    def p_pkg_orig_2(self, p):
        """pkg_orig : PKG_ORIG error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_ORIG_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_supplier_1(self, p):
        """pkg_supplier : PKG_SUPPL pkg_supplier_values"""
        try:
            self.builder.set_pkg_supplier(self.document, p[2])
        except OrderError:
            self.order_error("PackageSupplier", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageSupplier", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["PKG_SUPPL_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_pkg_supplier_2(self, p):
        """pkg_supplier : PKG_SUPPL error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_SUPPL_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_pkg_supplier_values_1(self, p):
        """pkg_supplier_values : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_pkg_supplier_values_2(self, p):
        """pkg_supplier_values : entity"""
        p[0] = p[1]

    def p_pkg_file_name(self, p):
        """pkg_file_name : PKG_FILE_NAME LINE"""
        try:
            value = p[2]
            self.builder.set_pkg_file_name(self.document, value)
        except OrderError:
            self.order_error("PackageFileName", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageFileName", p.lineno(1))

    def p_pkg_file_name_1(self, p):
        """pkg_file_name : PKG_FILE_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_FILE_NAME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_package_version_1(self, p):
        """package_version : PKG_VERSION LINE"""
        try:
            value = p[2]
            self.builder.set_pkg_vers(self.document, value)
        except OrderError:
            self.order_error("PackageVersion", "PackageName", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("PackageVersion", p.lineno(1))

    def p_package_version_2(self, p):
        """package_version : PKG_VERSION error"""
        self.error = True
        msg = ERROR_MESSAGES["PKG_VERSION_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_package_name(self, p):
        """package_name : PKG_NAME LINE"""
        try:
            value = p[2]
            self.builder.create_package(self.document, value)
            self.builder.set_current_package_name(value)
            self.builder.set_current_package_id(None)
            self.builder.set_current_file_name(None)  # start of a new package implies new file
            self.builder.set_current_file_id(None)
        except CardinalityError:
            self.more_than_one_error("PackageName", p.lineno(1))

    def p_package_name_1(self, p):
        """package_name : PKG_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["PACKAGE_NAME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_primary_package_purpose_1(self, p):
        """primary_package_purpose : PRIMARY_PACKAGE_PURPOSE primary_package_purpose_value"""
        try:
            self.builder.set_pkg_primary_package_purpose(self.document, p[2])
        except CardinalityError:
            self.more_than_one_error("PrimaryPackagePurpose", p.lineno(1))

    def p_primary_package_purpose_2(self, p):
        """primary_package_purpose : PRIMARY_PACKAGE_PURPOSE error"""
        self.error = True
        msg = ERROR_MESSAGES["PRIMARY_PACKAGE_PURPOSE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_primary_package_purpose_value(self, p):
        """primary_package_purpose_value : APPLICATION
                                         | FRAMEWORK
                                         | LIBRARY
                                         | CONTAINER
                                         | OPERATING_SYSTEM
                                         | DEVICE
                                         | FIRMWARE
                                         | SOURCE
                                         | ARCHIVE
                                         | FILE
                                         | INSTALL
                                         | OTHER
        """
        p[0] = p[1]

    def p_built_date_1(self, p):
        """built_date : BUILT_DATE DATE"""
        try:
            value = p[2]
            self.builder.set_pkg_built_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("BuiltDate", p.lineno(1))

    def p_built_date_2(self, p):
        """built_date : BUILT_DATE error"""
        self.error = True
        msg = ERROR_MESSAGES["BUILT_DATE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_release_date_1(self, p):
        """release_date : RELEASE_DATE DATE"""
        try:
            value = p[2]
            self.builder.set_pkg_release_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("ReleaseDate", p.lineno(1))

    def p_release_date_2(self, p):
        """release_date : RELEASE_DATE error"""
        self.error = True
        msg = ERROR_MESSAGES["RELEASE_DATE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_valid_until_date_1(self, p):
        """valid_until_date : VALID_UNTIL_DATE DATE"""
        try:
            value = p[2]
            self.builder.set_pkg_valid_until_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("ValidUntilDate", p.lineno(1))

    def p_valid_until_date_2(self, p):
        """valid_until_date : VALID_UNTIL_DATE error"""
        self.error = True
        msg = ERROR_MESSAGES["VALID_UNTIL_DATE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snip_spdx_id(self, p):
        """snip_spdx_id : SNIPPET_SPDX_ID LINE"""
        try:
            value = p[2]
            self.builder.create_snippet(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_SPDX_ID_VALUE"].format(p.lineno(2))
            self.logger.log(msg)

    def p_snip_spdx_id_1(self, p):
        """snip_spdx_id : SNIPPET_SPDX_ID error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_SPDX_ID_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_name(self, p):
        """snip_name : SNIPPET_NAME LINE"""
        try:
            value = p[2]
            self.builder.set_snippet_name(self.document, value)
        except OrderError:
            self.order_error("SnippetName", "SnippetSPDXID", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("SnippetName", p.lineno(1))

    def p_snippet_name_1(self, p):
        """snip_name : SNIPPET_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIPPET_NAME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_comment(self, p):
        """snip_comment : SNIPPET_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_snippet_comment(self.document, value)
        except OrderError:
            self.order_error("SnippetComment", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_COMMENT_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("SnippetComment", p.lineno(1))

    def p_snippet_comment_1(self, p):
        """snip_comment : SNIPPET_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_attribution_text_1(self, p):
        """snippet_attribution_text : SNIPPET_ATTRIBUTION_TEXT text_or_line"""
        try:
            value = p[2]
            self.builder.set_snippet_attribution_text(self.document, value)
        except CardinalityError:
            self.more_than_one_error("SnippetAttributionText", p.lineno(1))
        except OrderError:
            self.order_error(
                "SnippetAttributionText", "SnippetAttributionText", p.lineno(1)
            )

    def p_snippet_attribution_text_2(self, p):
        """snippet_attribution_text : SNIPPET_ATTRIBUTION_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIPPET_ATTRIBUTION_TEXT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_cr_text(self, p):
        """snip_cr_text : SNIPPET_CR_TEXT snip_cr_value"""
        try:
            self.builder.set_snippet_copyright(self.document, p[2])
        except OrderError:
            self.order_error("SnippetCopyrightText", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_COPYRIGHT_VALUE"].format(p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("SnippetCopyrightText", p.lineno(1))

    def p_snippet_cr_text_1(self, p):
        """snip_cr_text : SNIPPET_CR_TEXT error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_COPYRIGHT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_cr_value_1(self, p):
        """snip_cr_value : text_or_line"""
        p[0] = p[1]

    def p_snippet_cr_value_2(self, p):
        """snip_cr_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_snippet_cr_value_3(self, p):
        """snip_cr_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_snippet_lic_comment(self, p):
        """snip_lic_comment : SNIPPET_LICS_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_snippet_lic_comment(self.document, value)
        except OrderError:
            self.order_error("SnippetLicenseComments", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_LICS_COMMENT_VALUE"].format(p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("SnippetLicenseComments", p.lineno(1))

    def p_snippet_lic_comment_1(self, p):
        """snip_lic_comment : SNIPPET_LICS_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_LICS_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_text_or_line_value_1(self, p):
        """text_or_line : TEXT"""
        p[0] = p[1]

    def p_text_or_line_value_2(self, p):
        """text_or_line : LINE"""
        p[0] = p[1]

    def p_snip_from_file_spdxid(self, p):
        """snip_file_spdx_id : SNIPPET_FILE_SPDXID LINE"""
        try:
            value = p[2]
            self.builder.set_snip_from_file_spdxid(self.document, value)
        except OrderError:
            self.order_error("SnippetFromFileSPDXID", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_FILE_SPDXID_VALUE"].format(p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("SnippetFromFileSPDXID", p.lineno(1))

    def p_snip_from_file_spdxid_1(self, p):
        """snip_file_spdx_id : SNIPPET_FILE_SPDXID error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_FILE_SPDXID_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_concluded_license(self, p):
        """snip_lics_conc : SNIPPET_LICS_CONC conc_license"""
        try:
            self.builder.set_snip_concluded_license(self.document, p[2])
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_LICS_CONC_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except OrderError:
            self.order_error("SnippetLicenseConcluded", "SnippetSPDXID", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("SnippetLicenseConcluded", p.lineno(1))

    def p_snippet_concluded_license_1(self, p):
        """snip_lics_conc : SNIPPET_LICS_CONC error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_LICS_CONC_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_lics_info(self, p):
        """snip_lics_info : SNIPPET_LICS_INFO snip_lic_info_value"""
        try:
            self.builder.set_snippet_lics_info(self.document, p[2])
        except OrderError:
            self.order_error("LicenseInfoInSnippet", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["SNIP_LICS_INFO_VALUE"].format(p.lineno(1))
            self.logger.log(msg)

    def p_snippet_lics_info_1(self, p):
        """snip_lics_info : SNIPPET_LICS_INFO error"""
        self.error = True
        msg = ERROR_MESSAGES["SNIP_LICS_INFO_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_byte_range(self, p):
        """snip_byte_range : SNIPPET_BYTE_RANGE LINE"""
        try:
            self.builder.set_snippet_byte_range(self.document, p[2].strip())
        except OrderError:
            self.order_error("SnippetByteRange", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = "Value for Snippet ByteRange invalid in line {}.".format(p.lineno(1))
            self.logger.log(msg)

    def p_snippet_byte_range_1(self, p):
        """snip_byte_range : SNIPPET_BYTE_RANGE error"""

        self.error = True
        msg = "Reading of SnippetByteRange failed for line {}.".format(p.lineno(1))
        self.logger.log(msg)

    def p_snippet_line_range(self, p):
        """snip_line_range : SNIPPET_LINE_RANGE LINE"""
        try:
            self.builder.set_snippet_line_range(self.document, p[2].strip())
        except OrderError:
            self.order_error("SnippetLineRange", "SnippetSPDXID", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = "Value for Snippet LineRange invalid in line {}.".format(p.lineno(1))
            self.logger.log(msg)

    def p_snippet_line_range_1(self, p):
        """snip_line_range : SNIPPET_LINE_RANGE error"""

        self.error = True
        msg = "Reading of SnippetLineRange failed for line {}.".format(p.lineno(1))
        self.logger.log(msg)

    def p_snip_lic_info_value_1(self, p):
        """snip_lic_info_value : NONE"""
        p[0] = utils.SPDXNone()

    def p_snip_lic_info_value_2(self, p):
        """snip_lic_info_value : NO_ASSERT"""
        p[0] = utils.NoAssert()

    def p_snip_lic_info_value_3(self, p):
        """snip_lic_info_value : LINE"""
        value = p[1]
        p[0] = license.License.from_identifier(value)

    def p_reviewer_1(self, p):
        """reviewer : REVIEWER entity"""
        self.builder.add_reviewer(self.document, p[2])

    def p_reviewer_2(self, p):
        """reviewer : REVIEWER error"""
        self.error = True
        msg = ERROR_MESSAGES["REVIEWER_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_review_date_1(self, p):
        """review_date : REVIEW_DATE DATE"""
        try:
            value = p[2]
            self.builder.add_review_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("ReviewDate", p.lineno(1))
        except OrderError:
            self.order_error("ReviewDate", "Reviewer", p.lineno(1))

    def p_review_date_2(self, p):
        """review_date : REVIEW_DATE error"""
        self.error = True
        msg = ERROR_MESSAGES["REVIEW_DATE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_review_comment_1(self, p):
        """review_comment : REVIEW_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.add_review_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("ReviewComment", p.lineno(1))
        except OrderError:
            self.order_error("ReviewComment", "Reviewer", p.lineno(1))

    def p_review_comment_2(self, p):
        """review_comment : REVIEW_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["REVIEW_COMMENT_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_annotator_1(self, p):
        """annotator : ANNOTATOR entity"""
        self.builder.add_annotator(self.document, p[2])

    def p_annotator_2(self, p):
        """annotator : ANNOTATOR error"""
        self.error = True
        msg = ERROR_MESSAGES["ANNOTATOR_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_annotation_date_1(self, p):
        """annotation_date : ANNOTATION_DATE DATE"""
        try:
            value = p[2]
            self.builder.add_annotation_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("AnnotationDate", p.lineno(1))
        except OrderError:
            self.order_error("AnnotationDate", "Annotator", p.lineno(1))

    def p_annotation_date_2(self, p):
        """annotation_date : ANNOTATION_DATE error"""
        self.error = True
        msg = ERROR_MESSAGES["ANNOTATION_DATE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_annotation_comment_1(self, p):
        """annotation_comment : ANNOTATION_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.add_annotation_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("AnnotationComment", p.lineno(1))
        except OrderError:
            self.order_error("AnnotationComment", "Annotator", p.lineno(1))

    def p_annotation_comment_2(self, p):
        """annotation_comment : ANNOTATION_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["ANNOTATION_COMMENT_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_annotation_type_1(self, p):
        """annotation_type : ANNOTATION_TYPE annotation_type_value"""
        try:
            value = p[2]
            self.builder.add_annotation_type(self.document, value)
        except CardinalityError:
            self.more_than_one_error("AnnotationType", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["ANNOTATION_TYPE_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except OrderError:
            self.order_error("AnnotationType", "Annotator", p.lineno(1))

    def p_annotation_type_2(self, p):
        """annotation_type : ANNOTATION_TYPE error"""
        self.error = True
        msg = ERROR_MESSAGES["ANNOTATION_TYPE_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_annotation_spdx_id_1(self, p):
        """annotation_spdx_id : ANNOTATION_SPDX_ID LINE"""
        try:
            value = p[2]
            self.builder.set_annotation_spdx_id(self.document, value)
        except CardinalityError:
            self.more_than_one_error("SPDXREF", p.lineno(1))
        except OrderError:
            self.order_error("SPDXREF", "Annotator", p.lineno(1))

    def p_annotation_spdx_id_2(self, p):
        """annotation_spdx_id : ANNOTATION_SPDX_ID error"""
        self.error = True
        msg = ERROR_MESSAGES["ANNOTATION_SPDX_ID_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_relationship_1(self, p):
        """relationship : RELATIONSHIP relationship_value"""
        try:
            value = p[2]
            self.builder.add_relationship(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["RELATIONSHIP_VALUE"].format(p.lineno(1))
            self.logger.log(msg)
        except OrderError:
            self.order_error("Relationship_type", "Relationship", p.lineno(1))

    def p_relationship_2(self, p):
        """relationship : RELATIONSHIP error"""
        self.error = True
        msg = ERROR_MESSAGES["RELATIONSHIP_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_relationship_value_with_doc_ref(self, p):
        """relationship_value : DOC_REF_ID LINE"""
        p[0] = p[1] + ":" + p[2]

    def p_relationship_value_without_doc_ref(self, p):
        """relationship_value : LINE"""
        p[0] = p[1]

    def p_relationship_comment_1(self, p):
        """relationship_comment : RELATIONSHIP_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.add_relationship_comment(self.document, value)
        except OrderError:
            self.order_error("RelationshipComment", "Relationship", p.lineno(1))
        except CardinalityError:
            self.more_than_one_error("RelationshipComment", p.lineno(1))

    def p_relationship_comment_2(self, p):
        """relationship_comment : RELATIONSHIP_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["RELATIONSHIP_COMMENT_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_lics_list_ver_1(self, p):
        """locs_list_ver : LIC_LIST_VER LINE"""
        try:
            value = p[2]
            self.builder.set_lics_list_ver(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["LIC_LIST_VER_VALUE"].format(p[2], p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("LicenseListVersion", p.lineno(1))

    def p_lics_list_ver_2(self, p):
        """locs_list_ver : LIC_LIST_VER error"""
        self.error = True
        msg = ERROR_MESSAGES["LIC_LIST_VER_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_doc_comment_1(self, p):
        """doc_comment : DOC_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_doc_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("DocumentComment", p.lineno(1))

    def p_doc_comment_2(self, p):
        """doc_comment : DOC_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["DOC_COMMENT_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_doc_namespace_1(self, p):
        """doc_namespace : DOC_NAMESPACE LINE"""
        try:
            value = p[2]
            self.builder.set_doc_namespace(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["DOC_NAMESPACE_VALUE"].format(p[2], p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("DocumentNamespace", p.lineno(1))

    def p_doc_namespace_2(self, p):
        """doc_namespace : DOC_NAMESPACE error"""
        self.error = True
        msg = ERROR_MESSAGES["DOC_NAMESPACE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_data_license_1(self, p):
        """data_lics : DOC_LICENSE LINE"""
        try:
            value = p[2]
            self.builder.set_doc_data_lics(self.document, value)
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["DOC_LICENSE_VALUE"].format(p[2], p.lineno(2))
            self.logger.log(msg)
        except CardinalityError:
            self.more_than_one_error("DataLicense", p.lineno(1))

    def p_data_license_2(self, p):
        """data_lics : DOC_LICENSE error"""
        self.error = True
        msg = ERROR_MESSAGES["DOC_LICENSE_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_doc_name_1(self, p):
        """doc_name : DOC_NAME LINE"""
        try:
            value = p[2]
            self.builder.set_doc_name(self.document, value)
        except CardinalityError:
            self.more_than_one_error("DocumentName", p.lineno(1))

    def p_doc_name_2(self, p):
        """doc_name : DOC_NAME error"""
        self.error = True
        msg = ERROR_MESSAGES["DOC_NAME_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_ext_doc_refs_1(self, p):
        """ext_doc_ref : EXT_DOC_REF DOC_REF_ID DOC_URI EXT_DOC_REF_CHKSUM"""
        try:
            doc_ref_id = p[2]
            doc_uri = p[3]
            ext_doc_chksum = p[4]

            self.builder.add_ext_doc_refs(
                self.document, doc_ref_id, doc_uri, ext_doc_chksum
            )
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["EXT_DOC_REF_VALUE"].format(p.lineno(2))
            self.logger.log(msg)

    def p_ext_doc_refs_2(self, p):
        """ext_doc_ref : EXT_DOC_REF error"""
        self.error = True
        msg = ERROR_MESSAGES["EXT_DOC_REF_VALUE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_spdx_version_1(self, p):
        """spdx_version : DOC_VERSION LINE"""
        try:
            value = p[2]
            self.builder.set_doc_version(self.document, value)
        except CardinalityError:
            self.more_than_one_error("SPDXVersion", p.lineno(1))
        except SPDXValueError:
            self.error = True
            msg = ERROR_MESSAGES["DOC_VERSION_VALUE"].format(p[2], p.lineno(1))
            self.logger.log(msg)

    def p_spdx_version_2(self, p):
        """spdx_version : DOC_VERSION error"""
        self.error = True
        msg = ERROR_MESSAGES["DOC_VERSION_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_creator_comment_1(self, p):
        """creator_comment : CREATOR_COMMENT text_or_line"""
        try:
            value = p[2]
            self.builder.set_creation_comment(self.document, value)
        except CardinalityError:
            self.more_than_one_error("CreatorComment", p.lineno(1))

    def p_creator_comment_2(self, p):
        """creator_comment : CREATOR_COMMENT error"""
        self.error = True
        msg = ERROR_MESSAGES["CREATOR_COMMENT_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_creator_1(self, p):
        """creator : CREATOR entity"""
        self.builder.add_creator(self.document, p[2])

    def p_creator_2(self, p):
        """creator : CREATOR error"""
        self.error = True
        msg = ERROR_MESSAGES["CREATOR_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_created_1(self, p):
        """created : CREATED DATE"""
        try:
            value = p[2]
            self.builder.set_created_date(self.document, value)
        except CardinalityError:
            self.more_than_one_error("Created", p.lineno(1))

    def p_created_2(self, p):
        """created : CREATED error"""
        self.error = True
        msg = ERROR_MESSAGES["CREATED_VALUE_TYPE"].format(p.lineno(1))
        self.logger.log(msg)

    def p_entity_1(self, p):
        """entity : TOOL_VALUE
        """
        try:
            value = p[1]
            p[0] = self.builder.build_tool(self.document, value)
        except SPDXValueError:
            msg = ERROR_MESSAGES["TOOL_VALUE"].format(p[1], p.lineno(1))
            self.logger.log(msg)
            self.error = True
            p[0] = None

    def p_entity_2(self, p):
        """entity : ORG_VALUE
        """
        try:
            value = p[1]
            p[0] = self.builder.build_org(self.document, value)
        except SPDXValueError:
            msg = ERROR_MESSAGES["ORG_VALUE"].format(p[1], p.lineno(1))
            self.logger.log(msg)
            self.error = True
            p[0] = None

    def p_entity_3(self, p):
        """entity : PERSON_VALUE
        """
        try:
            value = p[1]
            p[0] = self.builder.build_person(self.document, value)
        except SPDXValueError:
            msg = ERROR_MESSAGES["PERSON_VALUE"].format(p[1], p.lineno(1))
            self.logger.log(msg)
            self.error = True
            p[0] = None

    def p_error(self, p):
        pass

    def build(self, **kwargs):
        self.lex = Lexer()
        self.lex.build(reflags=re.UNICODE)
        self.yacc = yacc.yacc(module=self, **kwargs)

    def parse(self, text):
        self.document = document.Document()
        self.error = False
        self.yacc.parse(text, lexer=self.lex)
        # FIXME: this state does not make sense
        self.builder.reset()
        validation_messages = ErrorMessages()
        # Report extra errors if self.error is False otherwise there will be
        # redundant messages
        validation_messages = self.document.validate(validation_messages)
        if not self.error:
            if validation_messages:
                for msg in validation_messages:
                    self.logger.log(msg)
                self.error = True
        return self.document, self.error



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from unittest import TestCase

import spdx.config as config
from spdx.creationinfo import CreationInfo
from spdx.creationinfo import Tool
from spdx.creationinfo import Organization
from spdx.creationinfo import Person
from spdx.version import Version


class TestCreationInfo(TestCase):
    maxDiff = None

    def test_timestamp(self):
        dt = datetime(2014, 4, 8, 13, 42, 12)
        ci_time = CreationInfo(created=dt)
        assert ci_time.created == dt

    def test_iso_format(self):
        dt = datetime(2014, 4, 8, 13, 42, 12)
        ci_time = CreationInfo(created=dt)
        assert ci_time.created_iso_format == "2014-04-08T13:42:12Z"

    def test_comment(self):
        ci_default = CreationInfo()
        assert ci_default.comment is None
        ci_comment = CreationInfo(comment='Hello There')
        assert ci_comment.comment == 'Hello There'

    def test_creators(self):
        ci = CreationInfo()
        assert len(ci.creators) == 0
        person = Person(name='Alice', email='alice@example.com')
        tool = Tool(name='spdxtron-9000')
        org = Organization(name='Acme', email='acme@example.com')
        ci.add_creator(tool)
        ci.add_creator(org)
        ci.add_creator(person)
        assert len(ci.creators) == 3
        assert tool in ci.creators
        assert org in ci.creators
        assert person in ci.creators
        ci.remove_creator(person)
        assert len(ci.creators) == 2
        assert tool in ci.creators
        assert org in ci.creators
        assert person not in ci.creators

    def test_license_list_version(self):
        ci = CreationInfo()
        assert ci.license_list_version == config.LICENSE_LIST_VERSION
        ci = CreationInfo(license_list_version=Version(major=1, minor=0))
        assert ci.license_list_version == Version(major=1, minor=0)
        assert not ci.license_list_version == Version(major=1, minor=2)



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rdflib import BNode
from rdflib import Graph
from rdflib import Literal
from rdflib import Namespace
from rdflib import RDF
from rdflib import RDFS
from rdflib import URIRef
from rdflib.compare import to_isomorphic

from spdx import config
from spdx import file
from spdx import license
from spdx import utils
from spdx.checksum import Checksum
from spdx.package import Package
from spdx.parsers.loggers import ErrorMessages
from spdx.relationship import Relationship
from spdx.utils import get_files_in_package
from spdx.writers.tagvalue import InvalidDocumentError

import warnings


class BaseWriter(object):
    """
    Base class for all Writer classes.
    Provide utility functions and stores shared fields.
    """

    def __init__(self, document, out):
        self.document = document
        self.out = out
        self.doap_namespace = Namespace("http://usefulinc.com/ns/doap#")
        self.spdx_namespace = Namespace("http://spdx.org/rdf/terms#")
        self.graph = Graph()

    def create_checksum_node(self, checksum: Checksum) -> BNode:
        """
        Return a node representing spdx.checksum.
        """
        algo = checksum.identifier.algorithm_to_rdf_representation() or 'checksumAlgorithm_sha1'
        checksum_node = BNode()
        type_triple = (checksum_node, RDF.type, self.spdx_namespace.Checksum)
        self.graph.add(type_triple)
        algorithm_triple = (
            checksum_node,
            self.spdx_namespace.algorithm,
            Literal('http://spdx.org/rdf/terms#' + algo),
        )
        self.graph.add(algorithm_triple)
        value_triple = (
            checksum_node,
            self.spdx_namespace.checksumValue,
            Literal(checksum.value),
        )
        self.graph.add(value_triple)
        return checksum_node

    def to_special_value(self, value):
        """
        Return proper spdx term or Literal
        """
        if isinstance(value, utils.NoAssert):
            return self.spdx_namespace.noassertion
        elif isinstance(value, utils.SPDXNone):
            return self.spdx_namespace.none
        else:
            return Literal(value)


class LicenseWriter(BaseWriter):
    """
    Handle all License classes from spdx.document module.
    """

    def __init__(self, document, out):
        super(LicenseWriter, self).__init__(document, out)

    def licenses_from_tree_helper(self, current, licenses):
        if isinstance(
            current, (license.LicenseConjunction, license.LicenseDisjunction)
        ):
            self.licenses_from_tree_helper(current.license_1, licenses)
            self.licenses_from_tree_helper(current.license_2, licenses)
        else:
            licenses.add(self.create_license_helper(current))

    def licenses_from_tree(self, tree):
        """
        Traverse conjunctions and disjunctions like trees and return a
        set of all licenses in it as nodes.
        """
        # FIXME: this is unordered!
        licenses = set()
        self.licenses_from_tree_helper(tree, licenses)
        return licenses

    def create_conjunction_node(self, conjunction):
        """
        Return a node representing a conjunction of licenses.
        """
        node = BNode()
        type_triple = (node, RDF.type, self.spdx_namespace.ConjunctiveLicenseSet)
        self.graph.add(type_triple)
        licenses = self.licenses_from_tree(conjunction)
        for lic in licenses:
            member_triple = (node, self.spdx_namespace.member, lic)
            self.graph.add(member_triple)
        return node

    def create_disjunction_node(self, disjunction):
        """
        Return a node representing a disjunction of licenses.
        """
        node = BNode()
        type_triple = (node, RDF.type, self.spdx_namespace.DisjunctiveLicenseSet)
        self.graph.add(type_triple)
        licenses = self.licenses_from_tree(disjunction)
        for lic in licenses:
            member_triple = (node, self.spdx_namespace.member, lic)
            self.graph.add(member_triple)
        return node

    def create_license_helper(self, lic):
        """
        Handle single(no conjunction/disjunction) licenses.
        Return the created node.
        """
        if isinstance(lic, license.ExtractedLicense):
            return self.create_extracted_license(lic)
        if lic.identifier.rstrip("+") in config.LICENSE_MAP:
            return URIRef(lic.url)
        else:
            matches = [
                l
                for l in self.document.extracted_licenses
                if l.identifier == lic.identifier
            ]
            if len(matches) != 0:
                return self.create_extracted_license(matches[0])
            else:
                lic = license.ExtractedLicense(lic.identifier)
                warnings.warn(
                    "Missing extracted license: {0}".format(lic.identifier)
                )
                return self.create_extracted_license(lic)

    def create_extracted_license(self, lic):
        """
        Handle extracted license.
        Return the license node.
        """
        licenses = list(
            self.graph.triples((None, self.spdx_namespace.licenseId, lic.identifier))
        )
        if len(licenses) != 0:
            return licenses[0][0]  # return subject in first triple
        else:
            license_node = BNode()
            type_triple = (
                license_node,
                RDF.type,
                self.spdx_namespace.ExtractedLicensingInfo,
            )
            self.graph.add(type_triple)
            ident_triple = (
                license_node,
                self.spdx_namespace.licenseId,
                Literal(lic.identifier),
            )
            self.graph.add(ident_triple)
            text_triple = (
                license_node,
                self.spdx_namespace.extractedText,
                Literal(lic.text),
            )
            self.graph.add(text_triple)
            if lic.full_name is not None:
                name_triple = (
                    license_node,
                    self.spdx_namespace.licenseName,
                    self.to_special_value(lic.full_name),
                )
                self.graph.add(name_triple)
            for ref in lic.cross_ref:
                triple = (license_node, RDFS.seeAlso, URIRef(ref))
                self.graph.add(triple)
            if lic.comment is not None:
                comment_triple = (license_node, RDFS.comment, Literal(lic.comment))
                self.graph.add(comment_triple)
            return license_node

    def create_license_node(self, lic):
        """
        Return a node representing a license.
        Could be a single license (extracted or part of license list.) or
        a conjunction/disjunction of licenses.
        """
        if isinstance(lic, license.LicenseConjunction):
            return self.create_conjunction_node(lic)
        elif isinstance(lic, license.LicenseDisjunction):
            return self.create_disjunction_node(lic)
        else:
            return self.create_license_helper(lic)

    def license_or_special(self, lic):
        """
        Check for special values spdx:none and spdx:noassertion.
        Return the term for the special value or the result of passing
        license to create_license_node.
        """
        if isinstance(lic, utils.NoAssert):
            return self.spdx_namespace.noassertion
        elif isinstance(lic, utils.SPDXNone):
            return self.spdx_namespace.none
        else:
            return self.create_license_node(lic)


class FileWriter(LicenseWriter):
    """
    Write spdx.file.File
    """

    FILE_TYPES = {
        file.FileType.SOURCE: "fileType_source",
        file.FileType.OTHER: "fileType_other",
        file.FileType.BINARY: "fileType_binary",
        file.FileType.ARCHIVE: "fileType_archive",
    }

    def __init__(self, document, out):
        super(FileWriter, self).__init__(document, out)

    def create_file_node(self, doc_file):
        """
        Create a node for spdx.file.
        """
        file_node = URIRef(
            "http://www.spdx.org/files#{id}".format(id=str(doc_file.spdx_id))
        )
        type_triple = (file_node, RDF.type, self.spdx_namespace.File)
        self.graph.add(type_triple)

        name_triple = (file_node, self.spdx_namespace.fileName, Literal(doc_file.name))
        self.graph.add(name_triple)

        if doc_file.has_optional_field("comment"):
            comment_triple = (file_node, RDFS.comment, Literal(doc_file.comment))
            self.graph.add(comment_triple)

        if doc_file.has_optional_field("file_types"):
            for f_type in doc_file.file_types:
                ftype = self.spdx_namespace[file.file_type_to_rdf(f_type)]
                ftype_triple = (file_node, self.spdx_namespace.fileType, ftype)
                self.graph.add(ftype_triple)

        for chk_sum in doc_file.checksums.values():
            self.graph.add(
                (
                    file_node,
                    self.spdx_namespace.checksum,
                    self.create_checksum_node(chk_sum),
                )
            )

        conc_lic_node = self.license_or_special(doc_file.conc_lics)
        conc_lic_triple = (
            file_node,
            self.spdx_namespace.licenseConcluded,
            conc_lic_node,
        )
        self.graph.add(conc_lic_triple)

        license_info_nodes = map(self.license_or_special, doc_file.licenses_in_file)
        for lic in license_info_nodes:
            triple = (file_node, self.spdx_namespace.licenseInfoInFile, lic)
            self.graph.add(triple)

        if doc_file.has_optional_field("license_comment"):
            comment_triple = (
                file_node,
                self.spdx_namespace.licenseComments,
                Literal(doc_file.license_comment),
            )
            self.graph.add(comment_triple)

        if doc_file.has_optional_field("attribution_text"):
            file_attribution_text_triple = (
                file_node,
                self.spdx_namespace.attributionText,
                Literal(doc_file.attribution_text),
            )
            self.graph.add(file_attribution_text_triple)

        cr_text_node = self.to_special_value(doc_file.copyright)
        cr_text_triple = (file_node, self.spdx_namespace.copyrightText, cr_text_node)
        self.graph.add(cr_text_triple)

        if doc_file.has_optional_field("notice"):
            notice_triple = (file_node, self.spdx_namespace.noticeText, doc_file.notice)
            self.graph.add(notice_triple)

        contrib_nodes = map(lambda c: Literal(c), doc_file.contributors)
        contrib_triples = [
            (file_node, self.spdx_namespace.fileContributor, node)
            for node in contrib_nodes
        ]
        for triple in contrib_triples:
            self.graph.add(triple)

        return file_node

    def files(self):
        """
        Return list of file nodes.
        """
        return map(self.create_file_node, self.document.files)

    def add_file_dependencies_helper(self, doc_file):
        """
        Handle dependencies for a single file.
        - doc_file - instance of spdx.file.File.
        """
        subj_triples = list(
            self.graph.triples(
                (None, self.spdx_namespace.fileName, Literal(doc_file.name))
            )
        )
        if len(subj_triples) != 1:
            raise InvalidDocumentError(
                "Could not find dependency subject {0}".format(doc_file.name)
            )
        subject_node = subj_triples[0][0]
        for dependency in doc_file.dependencies:
            dep_triples = list(
                self.graph.triples(
                    (None, self.spdx_namespace.fileName, Literal(dependency))
                )
            )
            if len(dep_triples) == 1:
                dep_node = dep_triples[0][0]
                dep_triple = (
                    subject_node,
                    self.spdx_namespace.fileDependency,
                    dep_node,
                )
                self.graph.add(dep_triple)
            else:
                print(
                    "Warning could not resolve file dependency {0} -> {1}".format(
                        doc_file.name, dependency
                    )
                )

    def add_file_dependencies(self):
        """
        Add file dependencies to the graph.
        Called after all files have been added.
        """
        for doc_file in self.document.files:
            self.add_file_dependencies_helper(doc_file)


class SnippetWriter(LicenseWriter):
    """
    Write spdx.snippet.Snippet
    """

    def __init__(self, document, out):
        super(SnippetWriter, self).__init__(document, out)

    def create_snippet_node(self, snippet):
        """
        Return a snippet node.
        """
        snippet_node = URIRef("http://spdx.org/rdf/terms/Snippet#" + snippet.spdx_id)
        type_triple = (snippet_node, RDF.type, self.spdx_namespace.Snippet)
        self.graph.add(type_triple)

        if snippet.has_optional_field("comment"):
            comment_triple = (snippet_node, RDFS.comment, Literal(snippet.comment))
            self.graph.add(comment_triple)

        if snippet.has_optional_field("name"):
            name_triple = (
                snippet_node,
                self.spdx_namespace.name,
                Literal(snippet.name),
            )
            self.graph.add(name_triple)

        if snippet.has_optional_field("license_comment"):
            lic_comment_triple = (
                snippet_node,
                self.spdx_namespace.licenseComments,
                Literal(snippet.license_comment),
            )
            self.graph.add(lic_comment_triple)

        if snippet.has_optional_field("attribution_text"):
            lic_attribution_text_triple = (
                snippet_node,
                self.spdx_namespace.attributionText,
                Literal(snippet.attribution_text),
            )
            self.graph.add(lic_attribution_text_triple)

        cr_text_node = self.to_special_value(snippet.copyright)
        cr_text_triple = (snippet_node, self.spdx_namespace.copyrightText, cr_text_node)
        self.graph.add(cr_text_triple)

        snip_from_file_triple = (
            snippet_node,
            self.spdx_namespace.snippetFromFile,
            Literal(snippet.snip_from_file_spdxid),
        )
        self.graph.add(snip_from_file_triple)

        conc_lic_node = self.license_or_special(snippet.conc_lics)
        conc_lic_triple = (
            snippet_node,
            self.spdx_namespace.licenseConcluded,
            conc_lic_node,
        )
        self.graph.add(conc_lic_triple)

        license_info_nodes = map(self.license_or_special, snippet.licenses_in_snippet)
        for lic in license_info_nodes:
            triple = (snippet_node, self.spdx_namespace.licenseInfoInSnippet, lic)
            self.graph.add(triple)

        return snippet_node

    def snippets(self):
        """
        Return list of snippet nodes.
        """
        return map(self.create_snippet_node, self.document.snippet)


class ReviewInfoWriter(BaseWriter):
    """
    Write spdx.review.Review
    """

    def __init__(self, document, out):
        super(ReviewInfoWriter, self).__init__(document, out)

    def create_review_node(self, review):
        """
        Return a review node.
        """
        review_node = BNode()
        type_triple = (review_node, RDF.type, self.spdx_namespace.Review)
        self.graph.add(type_triple)

        reviewer_node = Literal(review.reviewer.to_value())
        self.graph.add((review_node, self.spdx_namespace.reviewer, reviewer_node))
        reviewed_date_node = Literal(review.review_date_iso_format)
        reviewed_triple = (
            review_node,
            self.spdx_namespace.reviewDate,
            reviewed_date_node,
        )
        self.graph.add(reviewed_triple)
        if review.has_comment:
            comment_node = Literal(review.comment)
            comment_triple = (review_node, RDFS.comment, comment_node)
            self.graph.add(comment_triple)

        return review_node

    def reviews(self):
        "Returns a list of review nodes"
        return map(self.create_review_node, self.document.reviews)


class AnnotationInfoWriter(BaseWriter):
    """
    Write spdx.annotation.Annotation
    """

    def __init__(self, document, out):
        super(AnnotationInfoWriter, self).__init__(document, out)

    def create_annotation_node(self, annotation):
        """
        Return an annotation node.
        """
        annotation_node = URIRef(str(annotation.spdx_id))
        type_triple = (annotation_node, RDF.type, self.spdx_namespace.Annotation)
        self.graph.add(type_triple)

        annotator_node = Literal(annotation.annotator.to_value())
        self.graph.add((annotation_node, self.spdx_namespace.annotator, annotator_node))
        annotation_date_node = Literal(annotation.annotation_date_iso_format)
        annotation_triple = (
            annotation_node,
            self.spdx_namespace.annotationDate,
            annotation_date_node,
        )
        self.graph.add(annotation_triple)
        if annotation.has_comment:
            comment_node = Literal(annotation.comment)
            comment_triple = (annotation_node, RDFS.comment, comment_node)
            self.graph.add(comment_triple)
        annotation_type_node = Literal(annotation.annotation_type)
        annotation_type_triple = (
            annotation_node,
            self.spdx_namespace.annotationType,
            annotation_type_node,
        )
        self.graph.add(annotation_type_triple)

        return annotation_node

    def annotations(self):
        """
        Return a list of annotation nodes
        """
        return map(self.create_annotation_node, self.document.annotations)


class RelationshipInfoWriter(BaseWriter):
    """
    Write spdx.relationship.Relationship
    """

    def __init__(self, document, out):
        super(RelationshipInfoWriter, self).__init__(document, out)

    def transform_relationship_type_to_rdf_model(self, relationship_type: str) -> str:
        """
        Transform relationship type from upper snake case to camel case to match rdf-specific output e.g.
        COPY_OF -> relationshipType_copyOf.
        """
        return "relationshipType_" + relationship_type[0].lower() + relationship_type.title().replace('_', '')[1:]

    def create_relationship_node(self, relationship: Relationship) -> BNode:
        """
        Return a relationship node.
        """
        relationship_node = BNode()
        type_triple = (relationship_node, RDF.type, self.spdx_namespace.Relationship)
        self.graph.add(type_triple)

        relationship_spdx_element_id = Literal(relationship.spdx_element_id)
        self.graph.add(
            (
                relationship_node,
                self.spdx_namespace.spdxElementId,
                relationship_spdx_element_id,
            )
        )
        relationship_type = self.transform_relationship_type_to_rdf_model(relationship.relationship_type)
        relationship_type_node = self.spdx_namespace[relationship_type]
        self.graph.add(
            (
                relationship_node,
                self.spdx_namespace.relationshipType,
                relationship_type_node,
            )
        )
        related_spdx_node = Literal(relationship.related_spdx_element)
        related_spdx_triple = (
            relationship_node,
            self.spdx_namespace.relatedSpdxElement,
            related_spdx_node,
        )
        self.graph.add(related_spdx_triple)
        if relationship.has_comment:
            comment_node = Literal(relationship.relationship_comment)
            comment_triple = (relationship_node, RDFS.comment, comment_node)
            self.graph.add(comment_triple)

        return relationship_node

    def relationships(self):
        """
        Return a list of relationship nodes
        """
        return map(self.create_relationship_node, self.document.relationships)


class CreationInfoWriter(BaseWriter):
    """
    Write class spdx.creationinfo.CreationInfo
    """

    def __init__(self, document, out):
        super(CreationInfoWriter, self).__init__(document, out)

    def creators(self):
        """
        Return a list of creator nodes.
        Note: Does not add anything to the graph.
        """
        return map(
            lambda c: Literal(c.to_value()), self.document.creation_info.creators
        )

    def create_creation_info(self):
        """
        Add and return a creation info node to graph
        """
        ci_node = BNode()
        # Type property
        type_triple = (ci_node, RDF.type, self.spdx_namespace.CreationInfo)
        self.graph.add(type_triple)

        created_date = Literal(self.document.creation_info.created_iso_format)
        created_triple = (ci_node, self.spdx_namespace.created, created_date)
        self.graph.add(created_triple)

        creators = self.creators()
        for creator in creators:
            self.graph.add((ci_node, self.spdx_namespace.creator, creator))

        if self.document.creation_info.has_comment:
            comment_node = Literal(self.document.creation_info.comment)
            comment_triple = (ci_node, RDFS.comment, comment_node)
            self.graph.add(comment_triple)

        return ci_node


class ExternalDocumentRefWriter(BaseWriter):
    """
    Write class spdx.external_document_ref.ExternalDocumentRef
    """

    def __init__(self, document, out):
        super(ExternalDocumentRefWriter, self).__init__(document, out)

    def create_external_document_ref_node(self, ext_document_references):
        """
        Add and return a creation info node to graph
        """
        ext_doc_ref_node = BNode()
        type_triple = (
            ext_doc_ref_node,
            RDF.type,
            self.spdx_namespace.ExternalDocumentRef,
        )
        self.graph.add(type_triple)

        ext_doc_id = Literal(ext_document_references.external_document_id)
        ext_doc_id_triple = (
            ext_doc_ref_node,
            self.spdx_namespace.externalDocumentId,
            ext_doc_id,
        )
        self.graph.add(ext_doc_id_triple)

        doc_uri = Literal(ext_document_references.spdx_document_uri)
        doc_uri_triple = (ext_doc_ref_node, self.spdx_namespace.spdxDocument, doc_uri)
        self.graph.add(doc_uri_triple)

        checksum_node = self.create_checksum_node(ext_document_references.checksum)
        self.graph.add((ext_doc_ref_node, self.spdx_namespace.checksum, checksum_node))

        return ext_doc_ref_node

    def ext_doc_refs(self):
        """
        Return a list of review nodes
        """
        return map(
            self.create_external_document_ref_node,
            self.document.ext_document_references,
        )


class PackageWriter(LicenseWriter):
    """
    Write spdx.package.Package
    """

    def __init__(self, document, out):
        super(PackageWriter, self).__init__(document, out)

    def package_verif_node(self, package):
        """
        Return a node representing package verification code.
        """
        verif_node = BNode()
        type_triple = (
            verif_node,
            RDF.type,
            self.spdx_namespace.PackageVerificationCode,
        )
        self.graph.add(type_triple)
        value_triple = (
            verif_node,
            self.spdx_namespace.packageVerificationCodeValue,
            Literal(package.verif_code),
        )
        self.graph.add(value_triple)
        excl_file_nodes = map(lambda excl: Literal(excl), package.verif_exc_files)
        excl_predicate = self.spdx_namespace.packageVerificationCodeExcludedFile
        excl_file_triples = [
            (verif_node, excl_predicate, xcl_file) for xcl_file in excl_file_nodes
        ]
        for trp in excl_file_triples:
            self.graph.add(trp)
        return verif_node

    def handle_package_literal_optional(self, package, package_node, predicate, field):
        """
        Check if optional field is set.
        If so it adds the triple (package_node, predicate, $) to the graph.
        Where $ is a literal or special value term of the value of the field.
        """
        if package.has_optional_field(field):
            value = getattr(package, field, None)
            value_node = self.to_special_value(value)
            triple = (package_node, predicate, value_node)
            self.graph.add(triple)

    def handle_pkg_optional_fields(self, package: Package, package_node):
        """
        Write package optional fields.
        """
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.versionInfo, "version"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.packageFileName, "file_name"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.supplier, "supplier"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.originator, "originator"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.sourceInfo, "source_info"
        )
        self.handle_package_literal_optional(
            package,
            package_node,
            self.spdx_namespace.licenseComments,
            "license_comment",
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.summary, "summary"
        )
        self.handle_package_literal_optional(
            package,
            package_node,
            self.spdx_namespace.attributionText,
            "attribution_text",
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.description, "description"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.comment, "comment"
        )
        self.handle_package_literal_optional(
            package, package_node, self.spdx_namespace.filesAnalyzed, "files_analyzed"
        )

        if package.has_optional_field("checksums"):
            for checksum in package.checksums.values():
                checksum_node = self.create_checksum_node(checksum)
                self.graph.add((package_node, self.spdx_namespace.checksum, checksum_node))

        if package.has_optional_field("homepage"):
            homepage_node = URIRef(self.to_special_value(package.homepage))
            homepage_triple = (
                package_node,
                self.doap_namespace.homepage,
                homepage_node,
            )
            self.graph.add(homepage_triple)

    def create_package_node(self, package):
        """
        Return a Node representing the package.
        Files must have been added to the graph before this method is called.
        """
        package_node = URIRef(f"http://www.spdx.org/tools#{package.spdx_id}")
        type_triple = (package_node, RDF.type, self.spdx_namespace.Package)
        self.graph.add(type_triple)
        # Handle optional fields:
        self.handle_pkg_optional_fields(package, package_node)
        # package name
        name_triple = (package_node, self.spdx_namespace.name, Literal(package.name))
        self.graph.add(name_triple)
        # Package download location
        down_loc_node = (
            package_node,
            self.spdx_namespace.downloadLocation,
            self.to_special_value(package.download_location),
        )
        self.graph.add(down_loc_node)
        # Handle package verification
        if package.files_analyzed != False:
            verif_node = self.package_verif_node(package)
            verif_triple = (
                package_node,
                self.spdx_namespace.packageVerificationCode,
                verif_node,
            )
            self.graph.add(verif_triple)
        # Handle concluded license
        if package.conc_lics:
            conc_lic_node = self.license_or_special(package.conc_lics)
            conc_lic_triple = (
                package_node,
                self.spdx_namespace.licenseConcluded,
                conc_lic_node,
            )
            self.graph.add(conc_lic_triple)
        # Handle declared license
        if package.license_declared:
            decl_lic_node = self.license_or_special(package.license_declared)
            decl_lic_triple = (
                package_node,
                self.spdx_namespace.licenseDeclared,
                decl_lic_node,
            )
            self.graph.add(decl_lic_triple)
        # Package licenses from files
        licenses_from_files_nodes = map(
            lambda el: self.license_or_special(el), package.licenses_from_files
        )
        lic_from_files_predicate = self.spdx_namespace.licenseInfoFromFiles
        lic_from_files_triples = [
            (package_node, lic_from_files_predicate, node)
            for node in licenses_from_files_nodes
        ]
        for triple in lic_from_files_triples:
            self.graph.add(triple)
        # Copyright Text
        cr_text_node = self.to_special_value(package.cr_text)
        cr_text_triple = (package_node, self.spdx_namespace.copyrightText, cr_text_node)
        self.graph.add(cr_text_triple)
        # Handle files
        self.handle_package_has_file(package, package_node)
        return package_node

    def packages(self):
        """
        Return a node that represents the package in the graph.
        Call this function to write package info.
        """
        return [self.create_package_node(x)
                for x in self.document.packages]

    def handle_package_has_file_helper(self, pkg_file):
        """
        Return node representing pkg_file
        pkg_file should be instance of spdx.file.
        """
        nodes = list(
            self.graph.triples(
                (None, self.spdx_namespace.fileName, Literal(pkg_file.name))
            )
        )
        if len(nodes) == 1:
            return nodes[0][0]
        else:
            raise InvalidDocumentError(
                "handle_package_has_file_helper could not"
                + " find file node for file: {0}".format(pkg_file.name)
            )

    def handle_package_has_file(self, package, package_node):
        """
        Add hasFile triples to graph.
        Must be called after files have been added.
        """
        files = get_files_in_package(package, self.document.files, self.document.relationships)
        file_nodes = map(self.handle_package_has_file_helper, files)
        triples = [
            (package_node, self.spdx_namespace.hasFile, node) for node in file_nodes
        ]
        for triple in triples:
            self.graph.add(triple)


class PackageExternalRefWriter(BaseWriter):
    """
    Write class spdx.package.ExternalPackageRef
    """

    def __init__(self, document, out):
        super(PackageExternalRefWriter, self).__init__(document, out)

    def create_package_external_ref_node(self, pkg_ext_refs):
        """
        Add and return an external package reference node to graph.
        """
        pkg_ext_ref_node = BNode()
        pkg_ext_ref_triple = (
            pkg_ext_ref_node,
            RDF.type,
            self.spdx_namespace.ExternalRef,
        )
        self.graph.add(pkg_ext_ref_triple)

        pkg_ext_ref_category = Literal(pkg_ext_refs.category)
        pkg_ext_ref_category_triple = (
            pkg_ext_ref_node,
            self.spdx_namespace.referenceCategory,
            pkg_ext_ref_category,
        )
        self.graph.add(pkg_ext_ref_category_triple)

        pkg_ext_ref_type = Literal(pkg_ext_refs.pkg_ext_ref_type)
        pkg_ext_ref_type_triple = (
            pkg_ext_ref_node,
            self.spdx_namespace.referenceType,
            pkg_ext_ref_type,
        )
        self.graph.add(pkg_ext_ref_type_triple)

        pkg_ext_ref_locator = Literal(pkg_ext_refs.locator)
        pkg_ext_ref_locator_triple = (
            pkg_ext_ref_node,
            self.spdx_namespace.referenceLocator,
            pkg_ext_ref_locator,
        )
        self.graph.add(pkg_ext_ref_locator_triple)

        if pkg_ext_refs.comment:
            pkg_ext_ref_comment = Literal(pkg_ext_refs.comment)
            pkg_ext_ref_comment_triple = (
                pkg_ext_ref_node,
                RDFS.comment,
                pkg_ext_ref_comment,
            )
            self.graph.add(pkg_ext_ref_comment_triple)

        return pkg_ext_ref_node

    def pkg_ext_refs(self):
        """
        Return a list of package external references.
        """
        return [self.create_package_external_ref_node(ext_ref) for package in self.document.packages
                for ext_ref in package.pkg_ext_refs]


class Writer(
    CreationInfoWriter,
    ReviewInfoWriter,
    FileWriter,
    PackageWriter,
    PackageExternalRefWriter,
    ExternalDocumentRefWriter,
    AnnotationInfoWriter,
    RelationshipInfoWriter,
    SnippetWriter,
):
    """
    Wrapper for other writers to write all fields of spdx.document.Document
    Call `write()` to start writing.
    """

    def __init__(self, document, out):
        """
        - document is spdx.document instance that will be written.
        - out is a file-like object that will be written to.
        """
        super(Writer, self).__init__(document, out)

    def create_doc(self):
        """
        Add and return the root document node to graph.
        """
        doc_node = URIRef("http://www.spdx.org/tools#SPDXRef-DOCUMENT")
        # Doc type
        self.graph.add((doc_node, RDF.type, self.spdx_namespace.SpdxDocument))
        # Version
        vers_literal = Literal(str(self.document.version))
        self.graph.add((doc_node, self.spdx_namespace.specVersion, vers_literal))
        # Data license
        data_lics = URIRef(self.document.data_license.url)
        self.graph.add((doc_node, self.spdx_namespace.dataLicense, data_lics))
        if self.document.name:
            doc_name = URIRef(self.document.name)
            self.graph.add((doc_node, self.spdx_namespace.name, doc_name))
        return doc_node

    def write(self, doc_node=None):
        if not doc_node:
            doc_node = self.create_doc()
        # Add creation info
        creation_info_node = self.create_creation_info()
        ci_triple = (doc_node, self.spdx_namespace.creationInfo, creation_info_node)
        self.graph.add(ci_triple)
        # Add review info
        review_nodes = self.reviews()
        for review in review_nodes:
            self.graph.add((doc_node, self.spdx_namespace.reviewed, review))
        # Add external document references info
        ext_doc_ref_nodes = self.ext_doc_refs()
        for ext_doc_ref in ext_doc_ref_nodes:
            ext_doc_ref_triple = (
                doc_node,
                self.spdx_namespace.externalDocumentRef,
                ext_doc_ref,
            )
            self.graph.add(ext_doc_ref_triple)
        # Add extracted licenses
        licenses = map(self.create_extracted_license, self.document.extracted_licenses)
        for lic in licenses:
            self.graph.add(
                (doc_node, self.spdx_namespace.hasExtractedLicensingInfo, lic)
            )
        # Add files
        files = self.files()
        for file_node in files:
            self.graph.add((doc_node, self.spdx_namespace.referencesFile, file_node))
        self.add_file_dependencies()
        # Add package
        for package_node in self.packages():
            package_triple = (doc_node, self.spdx_namespace.describesPackage, package_node)
            self.graph.add(package_triple)

        # Add external package reference
        for pkg_ext_ref_node in self.pkg_ext_refs():
            pkg_ext_ref_triple = (
                doc_node,
                self.spdx_namespace.ExternalRef,
                pkg_ext_ref_node,
            )
            self.graph.add(pkg_ext_ref_triple)

        # Add relationship
        for relate_node in self.relationships():
            relate_triple = (doc_node, self.spdx_namespace.relationship, relate_node)
            self.graph.add(relate_triple)

        # Add snippet
        snippet_nodes = self.snippets()
        for snippet in snippet_nodes:
            self.graph.add((doc_node, self.spdx_namespace.Snippet, snippet))

        # normalize the graph to ensure that the sort order is stable
        self.graph = to_isomorphic(self.graph)

        # Write file
        self.graph.serialize(self.out, "pretty-xml", encoding="utf-8")


def write_document(document, out, validate=True):
    """
    Write an SPDX RDF document.
    - document - spdx.document instance.
    - out - file like object that will be written to.
    Optionally `validate` the document before writing and raise
    InvalidDocumentError if document.validate returns False.
    """

    if validate:
        messages = ErrorMessages()
        messages = document.validate(messages)
        if messages:
            raise InvalidDocumentError(messages)

    writer = Writer(document, out)
    writer.write()



# Copyright (c) Xavier Figueroa
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import io
import json
from unittest import TestCase

from spdx.parsers import jsonparser, yamlparser, xmlparser
from spdx.parsers.jsonyamlxmlbuilders import Builder
from spdx.parsers.loggers import StandardLogger

from tests import utils_test
from tests.utils_test import TestParserUtils


class TestParser(TestCase):
    maxDiff = None

    def check_document(self, document, expected_loc, regen=False):
        result = TestParserUtils.to_dict(document)

        if regen:
            with open(expected_loc, 'w') as o:
                o.write(json.dumps(result, indent=2))

        with io.open(expected_loc, encoding='utf-8') as ex:
            expected = json.load(ex, object_pairs_hook=OrderedDict)

        assert result == expected

    def test_json_parser(self):
        parser = jsonparser.Parser(Builder(), StandardLogger())
        test_file = utils_test.get_test_loc('formats/SPDXJsonExample.json')
        with io.open(test_file, encoding='utf-8') as f:
            document, _ = parser.parse(f)
        expected_loc = utils_test.get_test_loc('doc_parse/expected.json')
        self.check_document(document, expected_loc)

    def test_yaml_parser(self):
        parser = yamlparser.Parser(Builder(), StandardLogger())
        test_file = utils_test.get_test_loc('formats/SPDXYamlExample.yaml')
        with io.open(test_file, encoding='utf-8') as f:
            document, _ = parser.parse(f)
        expected_loc = utils_test.get_test_loc('doc_parse/expected.json')
        self.check_document(document, expected_loc)

    def test_xml_parser(self):
        parser = xmlparser.Parser(Builder(), StandardLogger())
        test_file = utils_test.get_test_loc('formats/SPDXXmlExample.xml')
        with io.open(test_file, encoding='utf-8') as f:
            document, _ = parser.parse(f)
        expected_loc = utils_test.get_test_loc('doc_parse/expected.json')
        self.check_document(document, expected_loc)

    def test_sbomyaml_parser(self):
        parser = yamlparser.Parser(Builder(), StandardLogger())
        test_file = utils_test.get_test_loc('formats/SPDXSBOMExample.spdx.yml')
        with io.open(test_file, encoding='utf-8') as f:
            document, errors = parser.parse(f)
            assert not errors
        expected_loc = utils_test.get_test_loc('doc_parse/SBOMexpected.json')
        self.check_document(document, expected_loc)



# Copyright (c) SPDX Python tools authors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from spdx.parsers import validations
from spdx import utils


class TestValidations(TestCase):
    maxDiff = None

    def test_validate_pkg_cr_text_does_not_crash_on_None(self):
        validations.validate_pkg_cr_text(None)

    def test_validate_pkg_cr_text_does_not_crash_on_NoAssert_or_SPDXNone(self):
        validations.validate_pkg_cr_text(utils.NoAssert())
        validations.validate_pkg_cr_text(utils.SPDXNone())

    def test_validate_file_cpyright_does_not_crash_on_None(self):
        validations.validate_file_cpyright(None)

    def test_validate_file_cpyright_does_not_crash_on_NoAssert_or_SPDXNone(self):
        validations.validate_file_cpyright(utils.NoAssert())
        validations.validate_file_cpyright(utils.SPDXNone())



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from functools import total_ordering

from spdx.utils import datetime_iso_format


@total_ordering
class Review(object):

    """
    Document review information.
    Fields:
    - reviewer: Person, Organization or tool that reviewed the SPDX file.
      Mandatory one.
    - review_date: Review date, mandatory one. Type: datetime.
    - comment: Review comment. Optional one. Type: str.
    """

    def __init__(self, reviewer=None, review_date=None, comment=None):
        self.reviewer = reviewer
        self.review_date = review_date
        self.comment = comment

    def __eq__(self, other):
        return (
            isinstance(other, Review)
            and self.reviewer == other.reviewer
            and self.review_date == other.review_date
            and self.comment == other.comment
        )

    def __lt__(self, other):
        return (self.reviewer, self.review_date, self.comment) < (
            other.reviewer,
            other.review_date,
            other.comment,
        )

    def set_review_date_now(self):
        self.review_date = datetime.utcnow().replace(microsecond=0)

    @property
    def review_date_iso_format(self):
        return datetime_iso_format(self.review_date)

    @property
    def has_comment(self):
        return self.comment is not None

    def validate(self, messages):
        """
        Check that all the fields are valid.
        Appends any error messages to messages parameter shall be a ErrorMessages.
        """
        self.validate_reviewer(messages)
        self.validate_review_date(messages)

        return messages

    def validate_reviewer(self, messages):
        if self.reviewer is None:
            messages.append("Review missing reviewer.")

        return messages

    def validate_review_date(self, messages):
        if self.review_date is None:
            messages.append("Review missing review date.")

        return messages



# Copyright (c) spdx contributors
# SPDX-License-Identifier: Apache-2.0
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from spdx.writers import json
from spdx.writers import yaml
from spdx.writers import rdf
from spdx.writers import xml
from spdx.writers import tagvalue
from spdx.parsers.builderexceptions import FileTypeError


def write_file(doc, fn, validate=True, encoding="utf-8"):
    out_mode = "w"
    if fn.endswith(".rdf") or fn.endswith(".rdf.xml"):
        writer_module = rdf
        out_mode = "wb"
        encoding = None
    elif fn.endswith(".tag") or fn.endswith(".spdx"):
        writer_module = tagvalue
    elif fn.endswith(".json"):
        writer_module = json
    elif fn.endswith(".xml"):
        writer_module = xml
    elif fn.endswith(".yaml"):
        writer_module = yaml
    else:
        raise FileTypeError("FileType Not Supported")

    with open(fn, out_mode, encoding=encoding) as out:
        p = writer_module.write_document(doc, out, validate)
